#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  csv2geojson.py
#
#         USAGE:  ./scripts/csv2geojson.py
#                 ./scripts/csv2geojson.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  Public Domain dedication or Zero-Clause BSD
#                 SPDX-License-Identifier: Unlicense OR 0BSD
#       VERSION:  v1.1.0
#       CREATED:  2023-04-11 18:13 BRT
#      REVISION:  2023-04-20 01:20 BRT v1.1 --contain-and-in
# ==============================================================================


# import geopandas
# import os
from abc import ABC, abstractmethod
import argparse
import csv
import json
import re
import sys
import string

from gisconflation.util import parse_argument_values


__VERSION__ = "1.1.0"
PROGRAM = "csv2geojson"
DESCRIPTION = """
------------------------------------------------------------------------------
CSV to GeoJSON

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
File on disk . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    {0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' \
--encoding='latin-1' data/tmp/DATASUS-tbEstabelecimento.csv

STDIN . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
(Note the "-" at the end)
    head data/tmp/DATASUS-tbEstabelecimento.csv | \
{0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' --encoding='latin-1' -


(With jq to format output)
    head data/tmp/DATASUS-tbEstabelecimento.csv | \
{0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' --encoding='latin-1' \
--ignore-warnings - | jq

GeoJSONSeq . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
    head data/tmp/DATASUS-tbEstabelecimento.csv | \
{0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' --encoding='latin-1' \
--output-type=GeoJSONSeq --ignore-warnings -

    head data/tmp/DATASUS-tbEstabelecimento.csv | \
{0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' --encoding='latin-1' \
--output-type=GeoJSONSeq --ignore-warnings - \
> data/tmp/DATASUS-tbEstabelecimento-head.geojsonl

GeoJSONSec -> Geopackage . . . . . . . . . . . . . . . . . . . . . . . . . . .
    {0} --lat=NU_LATITUDE --lon=NU_LONGITUDE --delimiter=';' --encoding='latin-1' \
--output-type=GeoJSONSeq --ignore-warnings \
data/tmp/DATASUS-tbEstabelecimento.csv \
> data/tmp/DATASUS-tbEstabelecimento.geojsonl

    ogr2ogr -f GPKG data/tmp/DATASUS-tbEstabelecimento.gpkg \
data/tmp/DATASUS-tbEstabelecimento.geojsonl

Exemplo CNPJ . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .
(https://www.gov.br/receitafederal/dados/cnpj-metadados.pdf)
CNPJ_BASICO;CNPJ_ORDEM;CNPJ_DV;IDENTIFICADOR;MATRIZ_FILIAL;SITUAÇÃO_CADASTRAL;DATA_SITUACAO_CADASTRAL;MOTIVO_SITUACAO_CADASTRAL;NOME_DA_CIDADE_NO_EXTERIOR;PAIS;DATA_DE_INICIO_ATIVIDADE;CNAE_FISCAL_PRINCIPAL;CNAE_FISCAL_SECUNDÁRIA;TIPO_DE_LOGRADOURO;LOGRADOURO;NUMERO;COMPLEMENTO;BAIRRO;CEP;UF;MUNICIPIO;DDD_1;TELEFONE_1;DDD_2;TELEFONE_2;DDD_DO_FAX;FAX;CORREIO_ELETRONICO;SITUACAO_ESPECIAL;DATA_DA_SITUACAO_ESPECIAL
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args

        Args:
            hxl_output (bool, optional): _description_. Defaults to True.
        """
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("input", help="path to CSV file on disk. Use - for stdin")

        parser.add_argument(
            "--lat",
            help="the name of the latitude column",
            dest="lat",
            required=True,
            nargs="?",
        )

        parser.add_argument(
            "--lon",
            help="the name of the longitude column",
            dest="lon",
            required=True,
            nargs="?",
        )

        parser.add_argument(
            "--filter-contain",
            help="Filter one or more fields for contain a string"
            "Use '|||' to divide the field and the string. "
            "Accept multiple values. "
            "Example: "
            "--filter-contain='name|||hospital'",
            dest="filter_contain",
            nargs="?",
            action="append",
        )

        parser.add_argument(
            "--filter-contain-regex",
            help="If defined, only results that match all clauses"
            " will appear on output. Accept multiple values."
            "Syntax is python regex. https://docs.python.org/3/library/re.html"
            "Example: "
            "--filter-contain-regex='name|||hospital.+'",
            dest="filter_contain_regex",
            nargs="?",
            action="append",
        )

        parser.add_argument(
            "--contain-or",
            help="If defined, only results that match at least one clause"
            " will appear on output. Accept multiple values."
            "--contain-or=tag1=value1 --contain-or=tag2=value2",
            dest="contain_or",
            nargs="?",
            action="append",
        )

        parser.add_argument(
            "--contain-and",
            help="If defined, only results that match all clauses"
            " will appear on output. Accept multiple values."
            "--contain-and=tag1=value1 --contain-and=tag2=value2",
            dest="contain_and",
            nargs="?",
            action="append",
        )

        parser.add_argument(
            "--contain-and-in",
            help="Alternative to -contain-and where values for a "
            "single field are a list. Separe values with ||"
            "Accept multiple values."
            "--contain-and=tag_a=valuea1||valuea2||valuea3",
            dest="contain_and_in",
            nargs="?",
            action="append",
        )

        parser.add_argument(
            "--delimiter",
            help="the type of delimiter",
            dest="delimiter",
            default=",",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--encoding",
            help="the type of delimiter",
            dest="encoding",
            default="utf-8",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-type",
            help="Change the default output type",
            dest="outfmt",
            default="GeoJSON",
            # geojsom
            # geojsonl
            choices=[
                "GeoJSON",
                "GeoJSONSeq",
            ],
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--ignore-warnings",
            help="Ignore some errors (such as empty latitude/longitude values)",
            dest="ignore_warnings",
            action="store_true",
        )

        cast_group = parser.add_argument_group(
            "Convert/preprocess data from input, including generate new fields"
        )

        cast_group.add_argument(
            "--cast-integer",
            help="Name of input fields to cast to integer. "
            "Use | for multiple. "
            "Example: <[ --cast-integer='field_a|field_b|field_c' ]>",
            dest="cast_integer",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        cast_group.add_argument(
            "--cast-float",
            help="Name of input fields to cast to float. "
            "Use | for multiple. "
            "Example: <[ --cast-float='latitude|longitude|field_c' ]>",
            dest="cast_float",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        cast_group.add_argument(
            "--column-copy-to",
            help="Add extra comluns. "
            "For multiple, use multiple times this parameter. "
            "Source vs destiny column must be divided by |. "
            "Example: <[ --column-copy-to='ORIGINAL_FIELD_PT|name:pt' "
            "--column-copy-to='CNPJ|ref:vatin' ]>",
            dest="column_copy",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        cast_group.add_argument(
            "--value-fixed",
            help="Define a fixed string for every value of a column, "
            "For multiple, use multiple times this parameter. "
            "Source vs destiny column must be divided by |. "
            "Example: <[ --value-fixed='source|BR:DATASUS' ]>",
            dest="value_fixed",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        cast_group.add_argument(
            "--value-prepend",
            help="Prepend a custom string to all values in a column. "
            "For multiple, use multiple times this parameter. "
            "Source vs destiny column must be divided by |. "
            "Example: <[ --value-prepend='ref:vatin|BR' ]>",
            dest="value_prepend",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        cast_group.add_argument(
            "--value-postcode-br",
            help="One or more column names to format as if was Brazilan postcodes, CEP",
            dest="value_postcode_br",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        cast_group.add_argument(
            "--value-phone-br",
            help="One or more column names to format as Brazilian "
            "phone/fax/WhatsApp number",
            dest="value_phone_br",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        cast_group.add_argument(
            "--value-name-place-br",
            help="One or more columsn to format as name of place (Locale BR)",
            dest="value_name_place_br",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        cast_group.add_argument(
            "--value-name-street-br",
            help="One or more columsn to format as name of street "
            "(Locale BR, 'logradouro') ",
            dest="value_name_street_br",
            nargs="?",
            type=lambda x: x.split("|"),
            default=None,
        )

        custom_group = parser.add_argument_group("Other")

        custom_group.add_argument(
            "--preprocessor-item-custom-inep",
            help="Custom feature not yet documented",
            dest="prepitem_custom_inep",
            required=False,
            nargs="?",
        )

        # _zzz_format_custom_cnefe
        custom_group.add_argument(
            "--preprocessor-complex-cnefe",
            help="Preprocessor for IBGE CNEFE 2022"
            "1: full metadata; 0: code-only; -1: no codes; -999: debug minimal",
            dest="prep_complex_cnefe",
            required=False,
            type=lambda x: int(x),
            nargs="?",
        )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # input_file = STDIN if pyargs.input == "-" else pyargs.input

        # print(parse_argument_values(
        #                 pyargs.filter_contain_regex
        #             ))

        # raise ValueError(pyargs)

        _contain_or = {}
        _contain_and = {}
        _contain_and_in = {}
        if pyargs.contain_or:
            for item in pyargs.contain_or:
                if item:
                    if item.find("="):
                        _key, _val = item.split("=")
                        _contain_or[_key] = _val
                    else:
                        _contain_or[_key] = True

        if pyargs.contain_and:
            for item in pyargs.contain_and:
                if item:
                    if item.find("="):
                        _key, _val = item.split("=")
                        _contain_and[_key] = _val
                    else:
                        _contain_and[_key] = True

        if pyargs.contain_and_in:
            for item in pyargs.contain_and_in:
                if item:
                    if item.find("="):
                        _key, _val = item.split("=")
                        _val = _val.strip("|").strip()
                        _val_items = _val.split("||")
                        _contain_and_in[_key] = _val_items
                    else:
                        raise SyntaxError("contain-and-in requires at least one value")

        with (
            open(pyargs.input, "r", encoding=pyargs.encoding)
            if len(pyargs.input) > 1
            else sys.stdin
        ) as csvfile:
            reader = csv.DictReader(csvfile, delimiter=pyargs.delimiter)

            if pyargs.outfmt == "GeoJSON":
                print('{"type": "FeatureCollection", "features": [')

            prepend = ""

            line_num = -1

            if pyargs.prep_complex_cnefe is not None:
                prep_cnefe = CustomProcessorCnefe()

            for row in reader:
                line_num += 1

                if not geojson_item_contain(
                    row,
                    contain_or=_contain_or,
                    contain_and=_contain_and,
                    contain_and_in=_contain_and_in,
                    filter_contain=parse_argument_values(pyargs.filter_contain),
                    filter_contain_regex=parse_argument_values(
                        pyargs.filter_contain_regex
                    ),
                ):
                    continue

                formated_row = row_item_cast(
                    row,
                    line_num=line_num,
                    cast_integer=pyargs.cast_integer,
                    cast_float=pyargs.cast_float,
                    ignore_warnings=pyargs.ignore_warnings,
                )

                if pyargs.prepitem_custom_inep:
                    formated_row = _zzz_format_custom_inep(
                        formated_row, pyargs.prepitem_custom_inep
                    )
                    # raise ValueError(pyargs.prepitem_custom_inep)

                if pyargs.prep_complex_cnefe is not None:
                    formated_row = prep_cnefe.item(
                        formated_row, pyargs.prep_complex_cnefe
                    )
                    # raise ValueError(pyargs.prep_complex_cnefe)

                row_v2 = row_item_column_add(
                    formated_row,
                    column_copy=pyargs.column_copy,
                    ignore_warnings=pyargs.ignore_warnings,
                )
                row_v3 = row_item_values(
                    row_v2,
                    value_fixed=pyargs.value_fixed,
                    value_prepend=pyargs.value_prepend,
                    value_postcode_br=pyargs.value_postcode_br,
                    value_phone_br=pyargs.value_phone_br,
                    value_name_place_br=pyargs.value_name_place_br,
                    value_name_street_br=pyargs.value_name_street_br,
                    ignore_warnings=pyargs.ignore_warnings,
                )

                item = geojson_item(
                    # formated_row,
                    # row_v2,
                    row_v3,
                    pyargs.lat,
                    pyargs.lon,
                    contain_or=_contain_or,
                    contain_and=_contain_and,
                    contain_and_in=_contain_and_in,
                    ignore_warnings=pyargs.ignore_warnings,
                )
                if not item:
                    continue

                jsonstr = json.dumps(item, ensure_ascii=False)

                # https://www.rfc-editor.org/rfc/rfc8142
                if pyargs.outfmt == "GeoJSONSeq":
                    print(f"\x1e{jsonstr}\n", sep="", end="")
                    continue

                print(f"{prepend} {jsonstr}")
                if prepend == "":
                    prepend = ","

            if pyargs.outfmt == "GeoJSON":
                print("]}")

        return self.EXIT_OK


def geojson_item(
    row,
    lat,
    lon,
    contain_or: list = None,
    contain_and: list = None,
    contain_and_in: list = None,
    ignore_warnings: bool = False,
):
    _lat = row[lat] if lat in row and len(row[lat].strip()) else False
    _lon = row[lon] if lon in row and len(row[lon].strip()) else False

    if not geojson_item_contain(
        row,
        contain_or=contain_or,
        contain_and=contain_and,
        contain_and_in=contain_and_in,
    ):
        return False

    if not _lat or not _lon:
        if not ignore_warnings:
            print(f"WARNING LAT/LON NOT FOUND [{row}]", file=sys.stderr)

        return False

    if _lat.find(",") > -1:
        _lat = _lat.replace(",", ".")

    if _lon.find(",") > -1:
        _lon = _lon.replace(",", ".")

    _lat = float(_lat)
    _lon = float(_lon)

    result = {
        "geometry": {"coordinates": [_lon, _lat], "type": "Point"},
        "properties": {},
        "type": "Feature",
    }

    _ignore = [lat, lon]

    result["properties"] = geojsom_item_properties(row, _ignore)

    return result


def geojson_item_contain(
    item,
    contain_or: list = None,
    contain_and: list = None,
    contain_and_in: list = None,
    filter_contain: list = None,
    filter_contain_regex: list = None,
) -> bool:
    if not item:
        return False

    if (
        not contain_or
        and not contain_and
        and not contain_and_in
        and not filter_contain_regex
    ):
        return True

    if contain_and:
        _count = len(contain_and.keys())

        for _key, _val in contain_and.items():
            if _key not in item:
                raise SyntaxError(f"key {_key} not in {item}")
                # return False

            if _val is not True and _val != item[_key]:
                return False
            _count -= 1

        if _count > 0:
            return False

    if contain_and_in:
        _count = len(contain_and_in.keys())

        for _key, _val in contain_and_in.items():
            if _key not in item:
                raise SyntaxError(f"key {_key} not in {item}")
                # return False

            if item[_key] not in _val:
                return False
            _count -= 1

        if _count > 0:
            return False

    for _key, _val in contain_or.items():
        if _key not in item:
            raise SyntaxError(f"key {_key} not in {item}")
            # return False

        if _val is not True and _val != item[_key]:
            return False

    if filter_contain:
        for _key, _val in filter_contain.items():
            _val = _val.lower()

            if _key not in item:
                raise SyntaxError(f"key {_key} not in {item}")
                # return False

            if _val is not True and item[_key].lower().find(_val) == -1:
                return False

    if filter_contain_regex:
        # raise ValueError(filter_contain_regex)
        for _key, _regex in filter_contain_regex.items():
            if _key not in item:
                raise SyntaxError(f"key {_key} not in {item}")
                # return False

            _result = re.match(_regex, item[_key])

            if not _result:
                return False

    return True


def geojsom_item_properties(row: dict, ignore: list):
    result = {}

    for key, value in row.items():
        if key in ignore:
            continue

        if not value or isinstance(value, str) and len(value.strip()) == 0:
            continue

        result[key] = value

    return result


def row_item_cast(
    row: dict,
    line_num: int = -1,
    cast_integer: list = None,
    cast_float: list = None,
    ignore_warnings: bool = False,
):
    result = {}

    for key, value in row.items():
        if isinstance(cast_integer, list) and key in cast_integer:
            if not value.isnumeric() and ignore_warnings:
                continue

            result[key] = int(value)

        elif isinstance(cast_float, list) and key in cast_float:
            if value.find(",") > -1:
                value = value.replace(",", ".")

            result[key] = float(value)
        else:
            result[key] = value

    return result


def row_item_column_add(
    row: dict,
    column_copy: list = None,
    cast_float: list = None,
    ignore_warnings: bool = False,
):
    _column_copy = {}
    if isinstance(column_copy, list) and len(column_copy) > 0:
        for item in column_copy:
            _key, _value = item.split("|")
            _column_copy[_key] = _value

    if len(_column_copy.keys()) == 0:
        return row

    result = {}
    result = row
    for key, value in _column_copy.items():
        if key not in row:
            if not ignore_warnings:
                # print("", file=sys.stderr)
                raise SyntaxError(f"row_item_column_add {key} not found")
        else:
            result[value] = result[key]

    #     elif isinstance(cast_float, list) and key in cast_float:
    #         if value.find(",") > -1:
    #             value = value.replace(",", ".")

    #         result[key] = float(value)
    #     else:
    #         result[key] = value

    return result


def row_item_values(
    row: dict,
    value_fixed: list = None,
    value_prepend: list = None,
    value_postcode_br: list = None,
    value_phone_br: list = None,
    value_name_place_br: list = None,
    value_name_street_br: list = None,
    ignore_warnings: bool = False,
):
    # result = row

    if isinstance(value_fixed, list) and len(value_fixed) > 0:
        for item in value_fixed:
            _key, _value = item.split("|")
            row[_key] = _value

    _value_prepend = {}
    if isinstance(value_prepend, list) and len(value_prepend) > 0:
        for item in value_prepend:
            _key, _value = item.split("|")
            _value_prepend[_key] = _value

    if len(_value_prepend.keys()) > 0:
        for key, value in _value_prepend.items():
            if key not in row:
                if not ignore_warnings:
                    # print("", file=sys.stderr)
                    raise SyntaxError(f"row_item_column_add {key} not found")
            else:
                # Sometimes the value is empty, so we don't prepend
                if row[key]:
                    row[key] = f"{value}{row[key]}"

    if isinstance(value_postcode_br, list) and len(value_postcode_br) > 0:
        for item in value_postcode_br:
            _value = _zzz_format_cep(row[item])
            if _value:
                row[item] = _value
            else:
                row[item] = ""

    if isinstance(value_phone_br, list) and len(value_phone_br) > 0:
        for item in value_phone_br:
            _value = _zzz_format_phone_br(row[item])
            if _value:
                row[item] = _value
            else:
                row[item] = ""

    if isinstance(value_name_place_br, list) and len(value_name_place_br) > 0:
        for item in value_name_place_br:
            _value = _zzz_format_name_place_br(row[item])
            if _value:
                row[item] = _value
            else:
                row[item] = ""

    if isinstance(value_name_street_br, list) and len(value_name_street_br) > 0:
        for item in value_name_street_br:
            _value = _zzz_format_name_street_br(row[item])
            if _value:
                row[item] = _value
            else:
                row[item] = ""

    return row


def _zzz_format_cep(value: str):
    if not value:
        return False
    if value.isnumeric():
        if len(value) == 8:
            return re.sub(r"(\d{5})(\d{3})", r"\1-\2", value)
    return False


def _zzz_format_name_place_br(value: str):
    if not value or not isinstance(value, str):
        return value

    term = string.capwords(value.strip())
    term2 = re.sub("\\s\\s+", " ", term)

    # @TODO deal with Do Da De

    return term2


def _zzz_format_name_street_br(value: str):
    if not value or not isinstance(value, str):
        return value

    term = string.capwords(value.strip())
    term2 = re.sub("\\s\\s+", " ", term)

    # @TODO deal with Do Da De
    # @TODO deal with abbreviations

    term2 = term2.replace(" Do ", " do ")
    term2 = term2.replace(" Dos ", " dos ")
    term2 = term2.replace(" Da ", " da ")
    term2 = term2.replace(" Das ", " das ")
    term2 = term2.replace(" De ", " de ")
    term2 = term2.replace(" Sao ", " São ")

    return term2


def _zzz_format_name_school_br(value: str):
    if not value or not isinstance(value, str):
        return value

    term = string.capwords(value.strip())
    term2 = re.sub("\\s\\s+", " ", term)

    # @TODO deal with Do Da De
    # @TODO deal with abbreviations

    return term2


def _zzz_format_phone_br(value: str):
    if not value:
        return False

    if value.startswith("+"):
        return value

    # @TODO deal with more than one number

    if value.startswith("(") and value.find(")") > -1:
        _num_com_ddd = "".join(filter(str.isdigit, value))

        _regiao = _num_com_ddd[0:2]
        _num_loc = _num_com_ddd[2:]
        _num_loc_p2 = _num_loc[-4:]
        _num_loc_p1 = _num_loc.replace(_num_loc_p2, "")
        # return "+55 " + _regiao + ' ' + _num_com_ddd[2:]
        return "+55 " + _regiao + " " + _num_loc_p1 + " " + _num_loc_p2

    # if value.isnumeric():
    #     if len(value) == 8:
    #         return re.sub(r"(\d{5})(\d{3})", r"\1-\2", value)
    return False


escolas_dict = {"ESC EST ENS FUN": "Escola Estadual Ensino Fundamental"}


# AVENIDA ALBERTO BINS, 410 5 ANDAR. CENTRO HISTORICO. 90030-140 Porto Alegre - RS.
#  - addr:floor=5
# https://pewu.github.io/osm-history/#/node/4163695342
# addr:floor
# pytest -vv src/csv2geojson/csv2geojson.py --doctest-modules
def _zzz_format_custom_inep(item: dict, source_column: str = "Endereço") -> dict:
    """_summary_

    Args:
        item (dict): _description_
        source_column (str, optional): _description_. Defaults to "Endereço".

    Returns:
        dict: _description_

    >>> d1 = "RUA LEONILLA HANSEN, 182 JOANETA. 95166-000 Picada Café - RS."
    >>> item1 = {"Endereço": d1}
    >>> r1 = _zzz_format_custom_inep(item1)
    >>> r1['addr:city']
    'Picada'
    >>> r1['addr:postcode']
    '95166-000'
    >>> d1 = "RUA DOS ANDRADAS, 1001 CONJUNTO 301. CENTRO HISTORICO. \
    ... 90020-015 Porto Alegre - RS."
    >>> item2 = {"Endereço": d1}
    >>> r2 = _zzz_format_custom_inep(item2)
    >>> r2['addr:housenumber']
    '1001'
    """
    result = item
    addr_raw = item[source_column]

    logradouro_arr = []
    parts = addr_raw.split(" ")
    while len(parts) > 0:
        token = parts.pop(0)
        # @TODO do the regex
        if len(token) == 9 and token[5] == "-":
            result["addr:postcode"] = token
            result["addr:city"] = parts.pop(0)
            break

        logradouro_arr.append(token)

    if addr_raw.find(", ") > -1:
        parts2 = addr_raw.split(", ")
        parts2b = parts2[1].split(" ")
        if parts2b[0].isnumeric():
            result["addr:street"] = _zzz_format_name_street_br(parts2[0])
            result["addr:housenumber"] = parts2b[0]

    # result["__addr:street"] = _zzz_format_name_street_br(
    #     " ".join(logradouro_arr).strip(".")
    # )
    # result["__addr:street"] = result["__addr:street"]p('.')

    return result


class CustomPreprocessorBase(ABC):
    @abstractmethod
    def item(self):
        pass


class CustomProcessorCnefe(CustomPreprocessorBase):

    COD_INDICADOR_FINALIDADE_CONST = {
        "1": "Residencial",
        "2": "Não residencial",
        "3": "Misto",
        "4": "Indeterminado",
    }

    COD_TIPO_ESPECIE = {
        "101": "Casa",
        "102": "Casa de vila ou em condomínio",
        "103": "Apartamento",
        "104": "Outros",
    }

    COD_ESPECIE = {
        "1": "Domicílio particular",
        "2": "Domicílio coletivo",
        "3": "Estabelecimento agropecuário",
        "4": "Estabelecimento de ensino",
        "5": "Estabelecimento de saúde",
        "6": "Estabelecimento de outras finalidades",
        "7": "Edificação em construção ou reforma",
        "8": "Estabelecimento religioso",
    }

    COD_INDICADOR_ESTAB_ENDERECO = {
        "1": "Único",
        "2": "Múltiplo, com até 10 estabelecimentos no endereço",
        "3": "Múltiplo, com mais de 10 estabelecimentos no endereço",
        "4": "Múltiplo, com quantidade de estabelecimentos desconhecida no endereço",
    }

    NV_GEO_COORD = {
        "1": "Endereço - coordenada original do Censo 2022",
        "2": "Endereço - coordenada modificada (apartamentos em um mesmo número no logradouro)",
        "3": "Endereço - coordenada estimada (endereços originalmente sem coordenadas ou coordenadas inválidas)",
        "4": "Face de quadra",
        "5": "Localidade",
        "6": "Setor censitário",
    }

    def item(self, item: dict, metadata: int = 1) -> dict:
        """Preprocessor for IBGE CNEFE

        @see https://ftp.ibge.gov.br/Cadastro_Nacional_de_Enderecos_para_Fins_Estatisticos/Censo_Demografico_2022/Arquivos_CNEFE/Dicionario_CNEFE_Censo_2022.xls
        Args:
            item (dict): _description_

        Returns:
            dict: _description_
        """
        result = item
        # addr_raw = item[source_column]

        # Versao baixada em 2024-05-29 tem erro em relacao ao dicionario
        # Aqui renomeia o campo que faltou um "E"
        if "COD_TIPO_ESPECI" in result:
            result["COD_TIPO_ESPECIE"] = result["COD_TIPO_ESPECI"]
            del result["COD_TIPO_ESPECI"]

        if "CEP" not in result or "NOM_SEGLOGR" not in result:
            raise KeyError("Bad file input. Is this CNEFE-like CSV?")

        # if "CEP" in result:
        result["addr:postcode"] = _zzz_format_cep(result["CEP"])
        del result["CEP"]

        logradouro = []
        logradouro.append(result["NOM_TIPO_SEGLOGR"])
        if result["NOM_TITULO_SEGLOGR"]:
            logradouro.append(result["NOM_TITULO_SEGLOGR"])

        logradouro.append(result["NOM_SEGLOGR"])

        if len(result["DSC_ESTABELECIMENTO"]) > 0:
            result["name"] = result["DSC_ESTABELECIMENTO"]

        if result["NOM_TIPO_SEGLOGR"] == "PRACA":
            result["addr:place"] = _zzz_format_name_street_br(" ".join(logradouro))
        else:
            result["addr:street"] = _zzz_format_name_street_br(" ".join(logradouro))

        if len(result["NUM_ENDERECO"]) > 0:
            # if result["NUM_ENDERECO"] == "S/N":
            if result["DSC_MODIFICADOR"] == "SN":
                result["addr:nohousenumber"] = "yes"
            else:
                result["addr:housenumber"] = result["NUM_ENDERECO"]

        # @TODO
        # VAL_COMP_ELEM1=CASA
        # VAL_COMP_ELEM1=1
        # CASA=1
        # addr:unit=Casa 1

        # result["source"] = "CNEFE 2022"
        result["source:addr"] = "CNEFE 2022"

        addr_extras = []
        if len(result["DSC_LOCALIDADE"]) > 0:
            addr_extras.append("localidade=" + result["DSC_LOCALIDADE"])

        if len(result["DSC_MODIFICADOR"]) > 0:
            addr_extras.append(
                result["DSC_MODIFICADOR"] + "=" + result["DSC_MODIFICADOR"]
            )

        if len(result["VAL_COMP_ELEM1"]) > 0:
            addr_extras.append(
                result["NOM_COMP_ELEM1"] + "=" + result["VAL_COMP_ELEM1"]
            )

        if len(result["VAL_COMP_ELEM2"]) > 0:
            addr_extras.append(
                result["NOM_COMP_ELEM2"] + "=" + result["VAL_COMP_ELEM2"]
            )

        if len(result["VAL_COMP_ELEM3"]) > 0:
            addr_extras.append(
                result["NOM_COMP_ELEM3"] + "=" + result["VAL_COMP_ELEM3"]
            )

        if len(result["VAL_COMP_ELEM4"]) > 0:
            addr_extras.append(
                result["NOM_COMP_ELEM4"] + "=" + result["VAL_COMP_ELEM4"]
            )

        if len(result["VAL_COMP_ELEM5"]) > 0:
            addr_extras.append(
                result["NOM_COMP_ELEM5"] + "=" + result["VAL_COMP_ELEM5"]
            )

        result["__meta_addr"] = ";".join(addr_extras)

        if metadata >= 1:
            if (
                result["COD_INDICADOR_FINALIDADE_CONST"]
                in self.COD_INDICADOR_FINALIDADE_CONST
            ):
                result["__meta_finalidade"] = self.COD_INDICADOR_FINALIDADE_CONST[
                    result["COD_INDICADOR_FINALIDADE_CONST"]
                ]

            if result["COD_TIPO_ESPECIE"] in self.COD_TIPO_ESPECIE:
                result["__meta_tipo_especie"] = self.COD_TIPO_ESPECIE[
                    result["COD_TIPO_ESPECIE"]
                ]

            if (
                result["COD_INDICADOR_ESTAB_ENDERECO"]
                in self.COD_INDICADOR_ESTAB_ENDERECO
            ):
                result["__meta_estab_endereco"] = self.COD_INDICADOR_ESTAB_ENDERECO[
                    result["COD_INDICADOR_ESTAB_ENDERECO"]
                ]

            if result["COD_ESPECIE"] in self.COD_ESPECIE:
                result["__meta_especie"] = self.COD_ESPECIE[result["COD_ESPECIE"]]

            if result["NV_GEO_COORD"] in self.NV_GEO_COORD:
                result["__meta_geo_coord"] = self.NV_GEO_COORD[result["NV_GEO_COORD"]]

        if metadata == 0:

            # _arr = [
            #     self.COD_INDICADOR_FINALIDADE_CONST[
            #         result["COD_INDICADOR_FINALIDADE_CONST"]
            #     ],
            #     self.COD_TIPO_ESPECIE[result["COD_TIPO_ESPECIE"]],
            #     self.COD_INDICADOR_ESTAB_ENDERECO[
            #         result["COD_INDICADOR_ESTAB_ENDERECO"]
            #     ],
            #     self.COD_ESPECIE[result["COD_ESPECIE"]],
            #     self.NV_GEO_COORD[result["NV_GEO_COORD"]],
            # ]

            _arr = []

            if (
                result["COD_INDICADOR_FINALIDADE_CONST"]
                in self.COD_INDICADOR_FINALIDADE_CONST
            ):
                _arr.append(result["COD_INDICADOR_FINALIDADE_CONST"])
            else:
                _arr.append("_")

            if result["COD_TIPO_ESPECIE"] in self.COD_TIPO_ESPECIE:
                _arr.append(result["COD_TIPO_ESPECIE"])
            else:
                _arr.append("_")

            if (
                result["COD_INDICADOR_ESTAB_ENDERECO"]
                in self.COD_INDICADOR_ESTAB_ENDERECO
            ):
                _arr.append(result["COD_INDICADOR_ESTAB_ENDERECO"])
            else:
                _arr.append("_")

            if result["COD_ESPECIE"] in self.COD_ESPECIE:
                _arr.append(result["COD_ESPECIE"])
            else:
                _arr.append("_")

            if result["NV_GEO_COORD"] in self.NV_GEO_COORD:
                _arr.append(result["NV_GEO_COORD"])

            else:
                _arr.append("_")

            result["__meta_codes"] = "-".join(_arr)

            # if (
            #     result["COD_INDICADOR_FINALIDADE_CONST"]
            #     in self.COD_INDICADOR_FINALIDADE_CONST
            # ):
            #     result["__meta_finalidade"] = self.COD_INDICADOR_FINALIDADE_CONST[
            #         result["COD_INDICADOR_FINALIDADE_CONST"]
            #     ]

            # if result["COD_TIPO_ESPECIE"] in self.COD_TIPO_ESPECIE:
            #     result["__meta_tipo_especie"] = self.COD_TIPO_ESPECIE[
            #         result["COD_TIPO_ESPECIE"]
            #     ]

            # if result["COD_INDICADOR_ESTAB_ENDERECO"] in self.COD_INDICADOR_ESTAB_ENDERECO:
            #     result["__meta_estab_endereco"] = self.COD_INDICADOR_ESTAB_ENDERECO[
            #         result["COD_INDICADOR_ESTAB_ENDERECO"]
            #     ]

            # if result["COD_ESPECIE"] in self.COD_ESPECIE:
            #     result["__meta_especie"] = self.COD_ESPECIE[result["COD_ESPECIE"]]

            # if result["NV_GEO_COORD"] in self.NV_GEO_COORD:
            #     result["__meta_geo_coord"] = self.NV_GEO_COORD[result["NV_GEO_COORD"]]

        # logradouro_arr = []
        # parts = addr_raw.split(" ")
        # while len(parts) > 0:
        #     token = parts.pop(0)
        #     # @TODO do the regex
        #     if len(token) == 9 and token[5] == "-":
        #         result["addr:postcode"] = token
        #         result["addr:city"] = parts.pop(0)
        #         break

        #     logradouro_arr.append(token)

        # if addr_raw.find(", ") > -1:
        #     parts2 = addr_raw.split(", ")
        #     parts2b = parts2[1].split(" ")
        #     if parts2b[0].isnumeric():
        #         result["addr:street"] = _zzz_format_name_street_br(parts2[0])
        #         result["addr:housenumber"] = parts2b[0]

        # # result["__addr:street"] = _zzz_format_name_street_br(
        # #     " ".join(logradouro_arr).strip(".")
        # # )
        # # result["__addr:street"] = result["__addr:street"]p('.')

        # delete other non-used fields

        if metadata == -999:
            del result["addr:postcode"]
            if "addr:street" in result:
                del result["addr:street"]
            del result["source:addr"]
            del result["__meta_addr"]

        del result["NOM_TIPO_SEGLOGR"]
        del result["NOM_TITULO_SEGLOGR"]
        del result["NOM_SEGLOGR"]

        del result["DSC_ESTABELECIMENTO"]

        del result["COD_UNICO_ENDERECO"]
        del result["COD_UF"]
        del result["COD_MUNICIPIO"]
        del result["COD_DISTRITO"]
        del result["COD_SUBDISTRITO"]
        del result["COD_SETOR"]
        del result["NUM_QUADRA"]
        del result["NUM_FACE"]
        del result["DSC_LOCALIDADE"]
        del result["NUM_ENDERECO"]
        del result["DSC_MODIFICADOR"]
        del result["NOM_COMP_ELEM1"]
        del result["VAL_COMP_ELEM1"]
        del result["NOM_COMP_ELEM2"]
        del result["VAL_COMP_ELEM2"]
        del result["NOM_COMP_ELEM3"]
        del result["VAL_COMP_ELEM3"]
        del result["NOM_COMP_ELEM4"]
        del result["VAL_COMP_ELEM4"]
        del result["NOM_COMP_ELEM5"]
        del result["VAL_COMP_ELEM5"]
        del result["COD_ESPECIE"]
        del result["NV_GEO_COORD"]
        del result["COD_INDICADOR_ESTAB_ENDERECO"]
        del result["COD_TIPO_ESPECIE"]
        del result["COD_INDICADOR_FINALIDADE_CONST"]
        del result["COD_INDICADOR_CONST_ENDERECO"]
        # del result["COD_UNICO_ENDERECO"]
        # del result["COD_UNICO_ENDERECO"]

        return result


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_2600 = Cli()
    args = cli_2600.make_args()
    # pyargs.print_help()

    # args.execute_cli(args)
    cli_2600.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  geojsonmerger.py
#
#         USAGE:  geojsonmerger --help
#                 ./src/gisconflation/geojsonmerger.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v0.1.0
#       CREATED:  2023-04-28 14:17 BRT
# ==============================================================================

import argparse
import csv

# import csv
# import dataclasses
import json
import re
import sys
from csv2geojson.csv2geojson import _zzz_format_phone_br

from gisconflation.util import AttributesEditor

# import logging
# from typing import List, Type

# from .util import AttributesEditor, parse_argument_values

# geojsonmerger tests/temp/dataset_a.geojson tests/temp/dataset_b.geojson
# geojsonmerger tests/temp/dataset_a.geojson tests/temp/dataset_b.geojson > tests/temp/dataset_a+b.geojson

# geojsonmerger tests/temp/dataset_a.geojson /workspace/git/fititnt/openstreetmap-vs-dados-abertos-brasil/data/tmp/DATASUS-tbEstabelecimento.csv
# geojsonmerger tests/temp/dataset_a.geojson /workspace/git/fititnt/openstreetmap-vs-dados-abertos-brasil/data/tmp/DATASUS-tbEstabelecimento.csv > tests/temp/dataset_csv_a+b.geojson
# geojsonmerger tests/temp/iede.rs.gov.br_Hospitais-no-RS_v2.geojson /workspace/git/fititnt/openstreetmap-vs-dados-abertos-brasil/data/tmp/DATASUS-tbEstabelecimento.csv > tests/temp/iede.rs.gov.br_Hospitais-no-RS_v2_plusmetadata.geojson

# other repo
# geojsonmerger data/tmp/DATASUS-tbEstabelecimento__RS.geojson data/tmp/ReceitaFederal_CNPJ_Estabelecimentos__RS_2023-05-09.csv
# geojsonmerger data/tmp/DATASUS-tbEstabelecimento__RS.geojson data/tmp/ReceitaFederal_CNPJ_Estabelecimentos__RS_2023-05-09.csv

__VERSION__ = "0.1.0"

PROGRAM = "geojsonmerger"
DESCRIPTION = """
------------------------------------------------------------------------------
[DRAFT] GeoJSON simple merger

Merge extra attributes from an GeoJSON (likely the one with more metadata, but
not accurate geometry) into another GeoJSON.

------------------------------------------------------------------------------
""".format(
    __file__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
    {0} --help

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer

# MATCH_EXACT = 1
# MATCH_NEAR = 3


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args

        Args:
            hxl_output (bool, optional): _description_. Defaults to True.
        """
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("geodataset_a", help="GeoJSON dataset 'A'")
        parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        # # @TODO maybe implement GeoJSONSeq
        # parser.add_argument("input", help="Input GeoJSON / GeoJSONSeq. Use - for stdin")
        # # parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        parser.add_argument(
            "--format-output",
            help="Path to output file",
            dest="format_output",
            default="auto",
            required=False,
            choices=["auto", "geojson", "geojsonseq"],
            nargs="?",
        )

        edit = parser.add_argument_group("Change properties from each item")

        edit.add_argument(
            "--rename-attribute",
            help="Rename attribute (if exist). "
            "Use '||' to divide the source key and target key. "
            "Accept multiple values. "
            "Example: "
            "--rename-attribute='NAME|||name' --rename-attribute='CITY|||addr:city'",
            dest="rename_attr",
            nargs="?",
            action="append",
        )

        # Old versions of csv2geojson uses only one |
        edit.add_argument(
            "--value-fixed",
            help="Define a fixed string for every value of a column, "
            "For multiple, use multiple times this parameter. "
            "Source vs destiny column must be divided by |||. "
            "Example: <[ --value-fixed='source|||BR:DATASUS' ]>",
            dest="value_fixed",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        edit.add_argument(
            "--value-norm-name-place",
            help="Column to normalize value (strategy: name of place). "
            "Accept multiple options. Example: "
            "--value-norm-name-place='name' --value-norm-name-place='alt_name'",
            dest="value_name_place",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        filter = parser.add_argument_group("Options for filter input items completely")

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # logger = logging.getLogger()
        # logger.setLevel(logging.INFO)
        # if pyargs.outlog:
        #     fh = logging.FileHandler(pyargs.outlog)
        #     logger.addHandler(fh)
        # else:
        #     ch = logging.StreamHandler()
        #     logger.addHandler(ch)

        # normalize_prop = True
        # skip_invalid_geometry = True

        # # raise ValueError(parse_argument_values(pyargs.value_fixed))
        # # print(parse_argument_values(pyargs.value_fixed))

        # gitem = GeoJSONItemEditor(
        #     rename_attr=parse_argument_values(pyargs.rename_attr),
        #     value_fixed=parse_argument_values(pyargs.value_fixed),
        #     value_name_place=pyargs.value_name_place,
        #     normalize_prop=normalize_prop,
        #     skip_invalid_geometry=skip_invalid_geometry,
        # )
        # gedit = GeoJSONFileEditor(pyargs.input, gitem)
        # gedit.output()

        # skip_invalid_geometry: True

        # # geodiff.debug()
        gmerger = GeoJSONMerger(
            geodataset_a=pyargs.geodataset_a,
            geodataset_b=pyargs.geodataset_b,
            # skip_invalid_geometry=True,
        )
        # print("TODO")

        gmerger.output()
        # gmerger.debug()
        return self.EXIT_OK


class GeoJSONMerger:
    def __init__(self, geodataset_a, geodataset_b) -> None:
        self.geodataset_a = geodataset_a
        self.geodataset_b = geodataset_b

        # @TODO make it flexible
        self.key_a = "ref:CNES"
        self.key_b = "CO_CNES"
        self.map_b = {
            "NU_CNPJ": "ref:vatin",
            "NU_CNPJ_MANTENEDORA": "operator:ref:vatin",
            "CO_CEP": "addr:postcode",
            "NO_EMAIL": "email",
            "NU_TELEFONE": "phone",
        }
        self.map_b_callback = {
            "NU_CNPJ": lambda x: f"BR{x}",
            "NU_CNPJ_MANTENEDORA": lambda x: f"BR{x}",
            "CO_CEP": lambda x: re.sub(r"(\d{5})(\d{3})", r"\1-\2", x),
            "NO_EMAIL": lambda x: x.lower(),
            "NU_TELEFONE": _zzz_format_phone_br,
        }

        self.in_a = []
        self.in_b = []
        self.out = []
        self.out_dict = {}

        self.statistics = {
            "a_items": 0,
            "a_items_valid": 0,
            "b_items": 0,
            "b_items_valid": 0,
            "ab_match": 0,
        }

        self._init_a()
        self._init_b()

    def _init_a(self) -> None:
        a_str = self._loader_temp(self.geodataset_a)

        a_data = json.loads(a_str)
        for item in a_data["features"]:
            self.statistics["a_items"] += 1
            if not "properties" in item or not item["properties"]:
                raise SyntaxError(item)

            if not self.key_a in item["properties"]:
                raise SyntaxError([self.key_a, item])

            key_active = str(item["properties"][self.key_a])
            if key_active in self.out_dict:
                # Improve this err handling
                print(f"Repeated value {key_active}")

            self.statistics["a_items_valid"] += 1
            self.out_dict[key_active] = item
            # print(item)

    def _init_b(self) -> None:
        # @TODO refactor this part
        if self.geodataset_b.lower().endswith((".csv", ".tsv", ".tab")):
            encoding = "latin-1"
            # _delimiter = get_delimiter(self.geodataset_b, encoding)
            _delimiter = ";"

            with open(self.geodataset_b, "r", encoding=encoding) as csvfile:
                reader = csv.DictReader(csvfile, delimiter=_delimiter)
                aedit = AttributesEditor(normalize_prop=True)
                for row in reader:
                    self.statistics["b_items"] += 1
                    if not self.key_b in row:
                        raise SyntaxError([self.key_b, item])

                    key_active = str(row[self.key_b])

                    self.statistics["b_items_valid"] += 1
                    if key_active not in self.out_dict:
                        continue
                    self.statistics["ab_match"] += 1

                    _props = self.out_dict[key_active]["properties"]
                    for _key, _val in row.items():
                        # @TODO allow some filtering
                        # _props[_key] = _val

                        if self.map_b_callback and _key in self.map_b_callback:
                            if _val:
                                _val = self.map_b_callback[_key](_val)

                        if self.map_b and _key in self.map_b:
                            _newkey = self.map_b[_key]
                            _props[_newkey] = _val
                        else:
                            _props[f"__b_{_key}"] = _val

                    _props2 = aedit.edit(_props)

                    # self.out_dict[key_active]["properties"] = _props
                    self.out_dict[key_active]["properties"] = _props2
                    # print(row)
                    # print(row['first_name'], row['last_name'])

            return True

        b_str = self._loader_temp(self.geodataset_b)

        b_data = json.loads(b_str)

        for item in b_data["features"]:
            self.statistics["b_items"] += 1
            if (
                not "properties" in item
                or not item["properties"]
                or not isinstance(item["properties"], dict)
            ):
                continue
                # raise SyntaxError(item)

            if not self.key_b in item["properties"]:
                raise SyntaxError([self.key_b, item])

            key_active = str(item["properties"][self.key_b])

            self.statistics["b_items_valid"] += 1

            if key_active not in self.out_dict:
                continue

            self.statistics["ab_match"] += 1
            # raise ValueError('deu')
            # print(key_active, type(key_active))
            # print(self.out_dict[key_active])
            # print('')
            # print(self.out_dict[key_active]["properties"])
            # print(item)

            _props = self.out_dict[key_active]["properties"]

            for _key, _val in item["properties"].items():
                # @TODO allow some filtering
                # _props[_key] = _val
                _props[f"__b_{_key}"] = _val

            # _props = {
            #     **self.out_dict[key_active]["properties"],
            #     **item["properties"][self.key_b],
            # }
            # self.out_dict[key_active]["properties"].update(
            #     item["properties"][self.key_b]
            # )
            self.out_dict[key_active]["properties"] = _props

    def _loader_temp(self, input_ptr: str):
        # @TODO make this efficient for very large files
        data = None
        if input_ptr == "-":
            temp = []
            for line in sys.stdin:
                # temp.append(line.strip())
                temp.append(line.rstrip())
                # temp.append(line)

            data = "\n".join(temp)
        else:
            with open(input_ptr, "r") as _file:
                data = _file.read()
            # pass

        return data

    def debug(self):
        # print("todo")
        print(self.statistics)

    def output(self):
        # @TODO keep other metadata at top level, if exist
        print('{"type": "FeatureCollection", "features": [')
        count = len(self.out_dict.keys())
        for _key, item in self.out_dict.items():
            count -= 1
            line_str = json.dumps(item, ensure_ascii=False)
            if count > 0:
                line_str += ","
            print(line_str)
        print("]}")

        # print("")
        # self.debug()


# https://stackoverflow.com/questions/3952132/how-do-you-dynamically-identify-unknown-delimiters-in-a-data-file
def get_delimiter(file_path: str, encoding="utf-8") -> str:
    with open(file_path, "r", encoding=encoding) as csvfile:
        delimiter = str(csv.Sniffer().sniff(csvfile.read()).delimiter)
        return delimiter


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  csvedit.py
#
#         USAGE:  ./scripts/csvedit.py
#                 ./scripts/csvedit.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v0.5.0
#       CREATED:  2023-05-06 16:54 BRT started, based on csv2geojson.py
#      REVISION:  --
# ==============================================================================


# import geopandas
# import os
import argparse
import csv

# import json
import re
import sys

from gisconflation.util import parse_argument_values

# import string

# from gisconflation.util import parse_argument_values

# try:
#     import geocoder
# except ImportError:
#     raise ImportError("Dependency not found. Please: pip install geocoder")
#
# # https://geocoder.readthedocs.io/providers/OpenStreetMap.html
# print(geocoder.osm('Mountain View, CA'))

# geocode 'New York city' --provider osm --out geojson | jq .

__VERSION__ = "0.5.0"
PROGRAM = "csvedit"
DESCRIPTION = """
------------------------------------------------------------------------------
CSV simple command line editor

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------

    {0} --help

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("input", help="path to CSV file on disk. Use - for stdin")

        # parser.add_argument(
        #     "--lat",
        #     help="the name of the latitude column",
        #     dest="lat",
        #     required=True,
        #     nargs="?",
        # )

        # parser.add_argument(
        #     "--lon",
        #     help="the name of the longitude column",
        #     dest="lon",
        #     required=True,
        #     nargs="?",
        # )

        # parser.add_argument(
        #     "--filter-contain",
        #     help="Filter one or more fields for contain a string"
        #     "Use '|||' to divide the field and the string. "
        #     "Accept multiple values. "
        #     "Example: "
        #     "--filter-contain='name|||hospital'",
        #     dest="filter_contain",
        #     nargs="?",
        #     action="append",
        # )

        # parser.add_argument(
        #     "--contain-or",
        #     help="If defined, only results that match at least one clause"
        #     " will appear on output. Accept multiple values."
        #     "--contain-or=tag1=value1 --contain-or=tag2=value2",
        #     dest="contain_or",
        #     nargs="?",
        #     action="append",
        # )

        # parser.add_argument(
        #     "--contain-and",
        #     help="If defined, only results that match all clauses"
        #     " will appear on output. Accept multiple values."
        #     "--contain-and=tag1=value1 --contain-and=tag2=value2",
        #     dest="contain_and",
        #     nargs="?",
        #     action="append",
        # )

        # parser.add_argument(
        #     "--contain-and-in",
        #     help="Alternative to -contain-and where values for a "
        #     "single field are a list. Separe values with ||"
        #     "Accept multiple values."
        #     "--contain-and=tag_a=valuea1||valuea2||valuea3",
        #     dest="contain_and_in",
        #     nargs="?",
        #     action="append",
        # )

        parser.add_argument(
            "--input-delimiter",
            help="The input delimiter",
            dest="in_delimiter",
            default=",",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--input-encoding",
            help="The input encoding",
            dest="in_encoding",
            default="utf-8",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--input-fieldnames",
            help="If the input CSV does not have a header, specify here. "
            "Use | as separator (if a field de de facto have |, then use \|). "
            "Example: --input-fieldnames='field with \| on it|another field'",
            dest="in_fieldnames",
            # default="utf-8",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-delimiter",
            help="The output delimiter",
            dest="out_delimiter",
            default=",",
            required=False,
            nargs="?",
        )

        filter_group = parser.add_argument_group("Filter rows")

        filter_group.add_argument(
            "--contain-and",
            help="If defined, only results that match all clauses"
            " will appear on output. Accept multiple values."
            "--contain-and=tag1=value1 --contain-and=tag2=value2",
            dest="contain_and",
            nargs="?",
            action="append",
        )

        filter_group.add_argument(
            "--contain-and-regex",
            help="If defined, only results that match all clauses"
            " will appear on output. Accept multiple values."
            "Syntax is python regex. https://docs.python.org/3/library/re.html"
            "Example: "
            "--contain-and-regex='name|||hospital.+'",
            dest="contain_and_regex",
            nargs="?",
            action="append",
        )

        filter_group.add_argument(
            "--filter-contain",
            help="Filter one or more fields for contain a string"
            "Use '|||' to divide the field and the string. "
            "Accept multiple values. "
            "Example: "
            "--filter-contain='name|||hospital'",
            dest="filter_contain",
            nargs="?",
            action="append",
        )

        # parser.add_argument(
        #     "--output-type",
        #     help="Change the default output type",
        #     dest="outfmt",
        #     default="GeoJSON",
        #     # geojsom
        #     # geojsonl
        #     choices=[
        #         "GeoJSON",
        #         "GeoJSONSeq",
        #     ],
        #     required=False,
        #     nargs="?",
        # )

        # parser.add_argument(
        #     "--ignore-warnings",
        #     help="Ignore some errors (such as empty latitude/longitude values)",
        #     dest="ignore_warnings",
        #     action="store_true",
        # )

        # cast_group = parser.add_argument_group(
        #     "Convert/preprocess data from input, including generate new fields"
        # )

        # cast_group.add_argument(
        #     "--cast-integer",
        #     help="Name of input fields to cast to integer. "
        #     "Use | for multiple. "
        #     "Example: <[ --cast-integer='field_a|field_b|field_c' ]>",
        #     dest="cast_integer",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--cast-float",
        #     help="Name of input fields to cast to float. "
        #     "Use | for multiple. "
        #     "Example: <[ --cast-float='latitude|longitude|field_c' ]>",
        #     dest="cast_float",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--column-copy-to",
        #     help="Add extra comluns. "
        #     "For multiple, use multiple times this parameter. "
        #     "Source vs destiny column must be divided by |. "
        #     "Example: <[ --column-copy-to='ORIGINAL_FIELD_PT|name:pt' "
        #     "--column-copy-to='CNPJ|ref:vatin' ]>",
        #     dest="column_copy",
        #     nargs="?",
        #     # type=lambda x: x.split("||"),
        #     action="append",
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-fixed",
        #     help="Define a fixed string for every value of a column, "
        #     "For multiple, use multiple times this parameter. "
        #     "Source vs destiny column must be divided by |. "
        #     "Example: <[ --value-fixed='source|BR:DATASUS' ]>",
        #     dest="value_fixed",
        #     nargs="?",
        #     # type=lambda x: x.split("||"),
        #     action="append",
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-prepend",
        #     help="Prepend a custom string to all values in a column. "
        #     "For multiple, use multiple times this parameter. "
        #     "Source vs destiny column must be divided by |. "
        #     "Example: <[ --value-prepend='ref:vatin|BR' ]>",
        #     dest="value_prepend",
        #     nargs="?",
        #     # type=lambda x: x.split("||"),
        #     action="append",
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-postcode-br",
        #     help="One or more column names to format as if was Brazilan postcodes, CEP",
        #     dest="value_postcode_br",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-phone-br",
        #     help="One or more column names to format as Brazilian "
        #     "phone/fax/WhatsApp number",
        #     dest="value_phone_br",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-name-place-br",
        #     help="One or more columsn to format as name of place (Locale BR)",
        #     dest="value_name_place_br",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        # cast_group.add_argument(
        #     "--value-name-street-br",
        #     help="One or more columsn to format as name of street "
        #     "(Locale BR, 'logradouro') ",
        #     dest="value_name_street_br",
        #     nargs="?",
        #     type=lambda x: x.split("|"),
        #     default=None,
        # )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # input_file = STDIN if pyargs.input == "-" else pyargs.input

        # print("TODO")

        _contain_and = {}
        if pyargs.contain_and:
            for item in pyargs.contain_and:
                if item:
                    if item.find("="):
                        _key, _val = item.split("=")
                        _contain_and[_key] = _val
                    else:
                        _contain_and[_key] = True

        filter_contain = parse_argument_values(pyargs.filter_contain)
        contain_and_regex = parse_argument_values(pyargs.contain_and_regex)

        # @TODO stdin does not yet allow non UTF8 customization (will pass as it is)
        # @see https://stackoverflow.com/questions/5004687
        with open(pyargs.input, "r", encoding=pyargs.in_encoding) if len(
            pyargs.input
        ) > 1 else sys.stdin as csvfile:
            if pyargs.in_fieldnames:
                _in_fieldnames = re.split(r"(?<!\\)\|", pyargs.in_fieldnames)
                # _in_fieldnames = re.split(r"", pyargs.in_fieldnames)
                # print(_in_fieldnames)

                reader = csv.DictReader(
                    csvfile, fieldnames=_in_fieldnames, delimiter=pyargs.in_delimiter
                )
            else:
                reader = csv.DictReader(csvfile, delimiter=pyargs.in_delimiter)

            # reader = csv.DictReader(csvfile, delimiter=pyargs.in_delimiter)

            firstline = next(reader)

            _fieldnames = firstline.keys()
            writer = csv.DictWriter(
                sys.stdout, fieldnames=_fieldnames, delimiter=pyargs.out_delimiter
            )

            writer.writeheader()
            # writer.writerow(firstline)

            reader2 = list(reader)
            reader2.insert(0, firstline)

            # @TODO bug with both conditions must be fixed.

            # for row in reader:
            for row in reader2:
                # @TODO move out of here
                if _contain_and:
                    _count = len(_contain_and.keys())

                    for _key, _val in _contain_and.items():
                        if _key not in row:
                            raise SyntaxError(f"key {_key} not in {row}")
                            # return False
                        # print(item)
                        if _val is not True and _val != row[_key]:
                            continue
                        _count -= 1

                    if _count > 0:
                        continue

                if filter_contain:
                    # raise ValueError(filter_contain)
                    _count2 = len(filter_contain.keys())

                    for _key, _val in filter_contain.items():
                        if not isinstance(_val, bool):
                            _val = _val.lower()

                        if _key not in row:
                            raise SyntaxError(f"key {_key} not in {row}")
                            # return False

                        if _val is not True and row[_key].lower().find(_val) == -1:
                            continue
                        _count2 -= 1

                    if _count2 > 0:
                        continue

                if contain_and_regex:
                    # raise ValueError(filter_contain)
                    _count3 = len(contain_and_regex.keys())

                    for _key, _regex in contain_and_regex.items():
                        # _val = _val.lower()

                        if _key not in row:
                            raise SyntaxError(f"key {_key} not in {row}")
                            # return False
                        _result = re.match(_regex, row[_key])

                        # if _val is not True and row[_key].lower().find(_val) == -1:
                        if not _result:
                            continue
                        _count3 -= 1

                    if _count3 > 0:
                        continue

                writer.writerow(row)

        return self.EXIT_OK


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_2600 = Cli()
    args = cli_2600.make_args()
    # pyargs.print_help()

    # args.execute_cli(args)
    cli_2600.execute_cli(args)



import re
import string


L10N_BR_LOWER = ["da", "de", "do"]


def normalize_place_name(value: str) -> str:
    """normalize_place_name Simplistic, prone to fail, name normalization

    @TODO allow localization

    Args:
        value (str): input term

    Returns:
        str: result term
    """
    if not value or not isinstance(value, str):
        return value

    term = string.capwords(value.strip())
    term2 = re.sub("\s\s+", " ", term)

    # @TODO deal with Do Da De

    return term2



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  json2csv.py
#
#         USAGE:  ./scripts/json2csv.py
#                 ./scripts/json2csv.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v1.0.0
#       CREATED:  2023-05-06 19:32 BRT
#      REVISION:  --
# ==============================================================================

import argparse
import csv
import json
import sys
from typing import Dict


__VERSION__ = "1.0.0"
PROGRAM = "json2csv"
DESCRIPTION = """
------------------------------------------------------------------------------
Convert JSON input file into CSV output file. If have nested values,
the output will contatenate the names with "."

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------

    {0} --help

    {0} input.json > output.csv

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

# Record separator
# DICTIONARY_SEPARATOR = "␞"

STDIN = sys.stdin.buffer


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("input", help="path to JSON file on disk. Use - for stdin")

        parser.add_argument(
            "--input-data-base",
            help="In case input data already is not a array of objects/list, "
            "this allow explicitly inform the base. Often is 'data'.",
            dest="in_base",
            default=None,  # Defaults to auto
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-delimiter",
            help="The input delimiter",
            dest="out_delimiter",
            default=",",
            required=False,
            nargs="?",
        )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        _checkfor = ["data"]
        data = False

        if pyargs.in_base:
            _checkfor.insert(0, pyargs.in_base)

        with open(pyargs.input, "r") if len(
            pyargs.input
        ) > 1 else sys.stdin as jsonfile:
            # parsed = json.loads(jsonfile)
            parsed = json.load(jsonfile)

            if isinstance(parsed, list):
                data = parsed
            elif isinstance(parsed, dict):
                for item in _checkfor:
                    if item in parsed:
                        data = parsed[item]
                        break

                if data is False:
                    raise SyntaxError(["Key not found. Options ", parsed.keys()])

            result_data = []

            fieldnames = []

            for item in data:
                _result = in_obiectum_planum(item)
                for _k in _result.keys():
                    if _k not in fieldnames:
                        fieldnames.append(_k)
                result_data.append(_result)

            writer = csv.DictWriter(
                sys.stdout, fieldnames=fieldnames, delimiter=pyargs.out_delimiter
            )

            writer.writeheader()

            for item in result_data:
                writer.writerow(item)

            return self.EXIT_OK


def in_obiectum_planum(rem: Dict, pydictsep: str = ".", pylistsep: str = " ") -> dict:
    """in_obiectum_planum Flatten a nested python object

    Trivia:
      - obiectum, https://en.wiktionary.org/wiki/obiectum#Latin
      - recursiōnem, https://en.wiktionary.org/wiki/recursio#Latin
      - praefīxum, https://en.wiktionary.org/wiki/praefixus#Latin
      - plānum, https://en.wiktionary.org/wiki/planus

    Args:
        rem (dict): The object to flatten
        pydictsep (pydictsep, optional): The separator for python dict keys
        pylistsep (pydictsep, optional): The separator for python list values

    Returns:
        [dict]: A flattened python dictionary

    Exemplōrum gratiā (et Python doctest, id est, testum automata):

    >>> testum1 = {'a0': {'a1': {'a2': 'va'}}, 'b0': [1, 2, 3]}
    >>> in_obiectum_planum(testum1)
    {'a0.a1.a2': 'va', 'b0': '1 2 3'}

    >>> in_obiectum_planum(testum1)
    {'a0.a1.a2': 'va', 'b0': '1 2 3'}

    >>> in_obiectum_planum(testum1, pylistsep=',')
    {'a0.a1.a2': 'va', 'b0': '1,2,3'}

    >>> in_obiectum_planum(testum1, pydictsep='->')
    {'a0->a1->a2': 'va', 'b0': '1 2 3'}

    # This is not designed to flat arrays, str, None, int, ..., only dict
    >>> in_obiectum_planum([1, 2, 3, 4])
    Traceback (most recent call last):
    ...
    TypeError: in_obiectum_planum non dict<class 'list'>
    """
    resultatum = {}

    if not isinstance(rem, dict):
        raise TypeError("in_obiectum_planum non dict" + str(type(rem)))

    def recursionem(rrem, praefixum: str = ""):
        praefixum_ad_hoc = "" if praefixum == "" else praefixum + pydictsep
        if isinstance(rrem, dict):
            for clavem in rrem:
                recursionem(rrem[clavem], praefixum_ad_hoc + clavem)
        elif isinstance(rrem, list):
            resultatum[praefixum] = pylistsep.join(map(str, rrem))
        else:
            resultatum[praefixum] = rrem

    recursionem(rem)

    return resultatum


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_2600 = Cli()
    args = cli_2600.make_args()
    # pyargs.print_help()

    # args.execute_cli(args)
    cli_2600.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  geojsondiff.py
#
#         USAGE:  geojsondiff --help
#                 ./src/gisconflation/geojsondiff.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#                   - haversine (pip install haversine)
#                   - shapely (pip install shapely)
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  Public Domain dedication or Zero-Clause BSD
#                 SPDX-License-Identifier: Unlicense OR 0BSD
#       VERSION:  v0.6.2
#       CREATED:  2023-04-16 22:36 BRT
#      REVISION:  2023-04-17 02:32 BRT v0.4.0 accept Overpas GeoJSON flavor
#                 2023-04-18 00:25 BRT v0.5.0 supports Polygon (not just Point)
#                 2023-04-19 21:52 BRT v0.6.0 draft of diff GeoJSON and JOSM
#                 2023-04-25 03:58 BRT v0.6.1 miggrade as pip package
#                 2023-06-03 03:16 BRT v0.6.2 moved to ./src/gisconflation
# ==============================================================================

import argparse
import csv
import dataclasses
import json
import sys
import logging
from typing import List, Type
import Levenshtein
from haversine import haversine, Unit

# from shapely.geometry import Polygon, Point
from shapely.geometry import Polygon
from xml.sax.saxutils import escape

from gisconflation.util import LevenshteinHelper


__VERSION__ = "0.6.2"
# VERSION = "0.6.2"

PROGRAM = "geojsondiff"
DESCRIPTION = """
------------------------------------------------------------------------------
GeoJSON++ diff v{1}

------------------------------------------------------------------------------
""".format(
    __file__, __VERSION__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
    {0} --output-diff-geojson=data/tmp/diff-points-ab.geojson \
--output-diff-tsv=data/tmp/diff-points-ab.tsv \
--output-diff-csv=data/tmp/diff-points-ab.csv \
--output-log=data/tmp/diff-points-ab.log.txt \
--tolerate-distance=1000 \
tests/data/data-points_a.geojson \
tests/data/data-points_b.geojson

GeoJSON (center point) example with overpass . . . . . . . . . . . . . . . . .
    [out:json][timeout:25];
    {{geocodeArea:Santa Catarina}}->.searchArea;
    (
    nwr["plant:source"="hydro"](area.searchArea);
    );
    convert item ::=::,::geom=geom(),_osm_type=type();
    out center;

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer

MATCH_EXACT = 1
MATCH_NEAR = 3


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args

        Args:
            hxl_output (bool, optional): _description_. Defaults to True.
        """
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument("geodataset_a", help="GeoJSON dataset 'A'")
        parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        parser.add_argument(
            "--output-diff-geojson",
            help="(Experimental) Path to output GeoJSON diff file",
            dest="outdiffgeo",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-diff-csv",
            help="Path to output CSV diff file",
            dest="outdiffcsv",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-diff-tsv",
            help="Path to output TSV (Tab-separated values) diff file",
            dest="outdifftsv",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--output-log",
            help="Path to output file",
            dest="outlog",
            default=None,
            required=False,
            nargs="?",
        )

        pivot = parser.add_argument_group(
            "Parameters used to know how to conflate A and B"
        )

        parser.add_argument(
            "--conflation-strategy",
            help="Conflation strategy.",
            dest="strategy",
            default="distance",
            required=False,
            choices=["distance", "addr"],
            nargs="?",
        )

        pivot.add_argument(
            "--tolerate-distance",
            help="Typical maximum distance for features match if not "
            "exact same point. In meters. Default to 100",
            dest="tdist",
            default="100",
            required=False,
            nargs="?",
        )

        pivot.add_argument(
            "--pivot-key-main",
            help="If defined, its an strong hint that item from A and B "
            "alredy are mached with each other. "
            "Use '||' if attribute on A is not the same on the B. "
            "Accept multiple values. "
            "Example: "
            "--pivot-key-main='CO_CNES||ref:CNES' --pivot-key-main='ref:vatin'",
            dest="pivot_key_main",
            nargs="?",
            action="append",
        )

        pivot.add_argument(
            "--pivot-attr-2",
            help="A non primary attribute on A and B (like phone or website) "
            "which while imperfect, is additional hint about being about same. "
            "Use '||' if attribute on A is not the same on the B. "
            "Accept multiple values. "
            "Example: "
            "--pivot-attr-2='contact:email'",
            dest="pivot_attr_2",
            nargs="?",
            action="append",
        )

        pivot.add_argument(
            "--pivot-alias",
            help="A weak (like name or description) attribute to order the matches. "
            "Impefect. Never will return perfect match, even if exact the same. "
            "Use '||' if attribute on A is not the same on the B. "
            "Accept multiple values. "
            "Example: "
            "--pivot-alias'COMPANY_NAME||name' --pivot-alias'COMPANY_NAME||alt_name'",
            dest="pivot_alias",
            nargs="?",
            action="append",
        )

        prefilter = parser.add_argument_group("Pre-filter data before processing")

        prefilter.add_argument(
            "--prefilter-a-contain",
            help="Prefilter (e.g. before processing) a filed in A for a string"
            "Use '||' to divide the field and the string. "
            "Accept multiple values. "
            "Example: "
            "--prefilter-a-contain='NO_RAZAO_SOCIAL||hospital'",
            dest="prefilter_a_contain",
            nargs="?",
            action="append",
        )

        prefilter.add_argument(
            "--prefilter-b-contain",
            help="Prefilter (e.g. before processing) a filed in B for a string"
            "Use '||' to divide the field and the string. "
            "Accept multiple values. "
            "Example: "
            "--prefilter-a-contain='name||hospital'",
            dest="prefilter_b_contain",
            nargs="?",
            action="append",
        )

        filters = parser.add_argument_group(
            "Quick output filters for GeoJSON output. Ignored by tabular diffs"
        )

        filters.add_argument(
            "--filter-ab-dist-min",
            help="Minimal distance between A and B. Example: 0",
            dest="filter_ab_dist_min",
            default=None,
            required=False,
            nargs="?",
        )

        filters.add_argument(
            "--filter-ab-dist-max",
            help="Minimal distance between A and B. Example: 0",
            dest="filter_ab_dist_max",
            default=None,
            required=False,
            action="store_true",
            # nargs="?",
        )

        filters.add_argument(
            "--filter-matched-pivot-key",
            help="Only output items matched by main pivot key",
            dest="filter_matched_key",
            default=None,
            required=False,
            action="store_true",
        )

        filters.add_argument(
            "--filter-matched-pivot-key-not",
            help="Do not output items matched by main pivot key",
            dest="filter_matched_key_not",
            default=None,
            required=False,
            action="store_true",
        )

        advanced = parser.add_argument_group(
            "ADVANCED. Do not upload to OpenStreetMap unless you review the output. "
            "Requires A be an external dataset, and B be OpenStreetMap data. "
        )

        advanced.add_argument(
            "--output-josm-file",
            help="Output OpenStreetMap change proposal in JOSM file format",
            dest="outosc",
            # default="100",s
            required=False,
            nargs="?",
        )

        # parser.add_argument(
        #     "--tolerate-distance-extra",
        #     help="Path to output file",
        #     dest="tdist",
        #     default="500",
        #     required=False,
        #     nargs="?",
        # )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        if pyargs.outlog:
            fh = logging.FileHandler(pyargs.outlog)
            logger.addHandler(fh)
        else:
            ch = logging.StreamHandler()
            logger.addHandler(ch)

        # distance_okay = 50
        # distance_okay = int(pyargs.tdist)
        # distance_permissive = 250

        # @TODO actually allow multiple primary keys or remove documentation
        pks_dict = parse_argument_values(pyargs.pivot_key_main)
        pivot_key_main = None
        if len(pks_dict.items()) > 0:
            _k = list(pks_dict.keys())[0]
            pivot_key_main = pks_dict[_k]

        # print('oi', pyargs.pivot_key_main, pks_dict)

        crules = ConflationRules(
            distance_okay=int(pyargs.tdist),
            # pivot_key_main=parse_argument_values(pyargs.pivot_key_main),
            pivot_key_main=pivot_key_main,
            pivot_attr_2=parse_argument_values(pyargs.pivot_attr_2),
            pivot_alias=parse_argument_values(pyargs.pivot_alias),
        )

        cfilters = ConflationFilters(
            filter_ab_dist_min=pyargs.filter_ab_dist_min,
            filter_ab_dist_max=pyargs.filter_ab_dist_max,
            filter_matched_key=pyargs.filter_matched_key,
            filter_matched_key_not=pyargs.filter_matched_key_not,
        )

        cprefilters = ConflationPrefilters(
            prefilter_a_contain=parse_argument_values(pyargs.prefilter_a_contain),
            prefilter_b_contain=parse_argument_values(pyargs.prefilter_b_contain),
        )

        geodiff = GeojsonCompare(
            pyargs.geodataset_a,
            pyargs.geodataset_b,
            pyargs.strategy,
            crules,
            cprefilters=cprefilters,
            cfilters=cfilters,
            logger=logger,
        )

        if pyargs.outosc:
            with open(pyargs.outosc, "w") as file:
                file.write(geodiff.osmchange())

        if pyargs.outdiffcsv:
            with open(pyargs.outdiffcsv, "w") as file:
                tabular_writer(file, geodiff.summary_tabular(), delimiter=",")

        if pyargs.outdifftsv:
            with open(pyargs.outdifftsv, "w") as file:
                tabular_writer(file, geodiff.summary_tabular(), delimiter="\t")

        if pyargs.outdiffgeo:
            with open(pyargs.outdiffgeo, "w") as file:
                geojson_diff = geodiff.diff_geojson_full()
                file.write(json.dumps(geojson_diff, ensure_ascii=False, indent=2))
                # tabular_writer(file, geodiff.summary_tabular(), delimiter="\t")

        # geodiff.debug()
        return self.EXIT_OK


class ConflationFilters:
    def __init__(
        self,
        filter_ab_dist_min: int = None,
        filter_ab_dist_max: int = None,
        filter_matched_key: bool = None,
        filter_matched_key_not: bool = None,
    ) -> None:
        if filter_ab_dist_min is not None:
            self.filter_ab_dist_min = float(filter_ab_dist_min)
        else:
            self.filter_ab_dist_min = None

        if filter_ab_dist_max is not None:
            self.filter_ab_dist_max = float(filter_ab_dist_max)
        else:
            self.filter_ab_dist_max = None

        if filter_matched_key is not None:
            self.filter_matched_key = filter_matched_key
        else:
            self.filter_matched_key = None

        if filter_matched_key_not is not None:
            self.filter_matched_key_not = filter_matched_key_not
        else:
            self.filter_matched_key_not = None

    def dist_ab(self, item_dist=int):
        if item_dist is None:
            return False
        if (
            self.filter_ab_dist_min is not None
            and not item_dist >= self.filter_ab_dist_min
        ):
            return False
        if (
            self.filter_ab_dist_max is not None
            and not item_dist <= self.filter_ab_dist_max
        ):
            return False

        return True

    def pk_allow(
        self,
        item_index: int,
        current_dataset: Type["DatasetInMemory"],
        other_dataset: Type["DatasetInMemory"],
    ):
        if self.filter_matched_key_not is None and self.filter_matched_key is None:
            return True

        if (
            self.filter_matched_key_not is not None
            and self.filter_matched_key is not None
        ):
            raise SyntaxError("Conflict: filter_matched_key + filter_matched_key_not ")

        item_pk = current_dataset.item_pk(item_index)
        # print(">> o", item_pk, other_dataset.pivot_keys_in)
        # print("<<c", item_pk, current_dataset.pivot_keys_in)
        # print(current_dataset.pivot_keys_in)
        # print(current_dataset.pivot_keys_in_duplicate)
        if (
            item_pk is not None
            and (
                self.filter_matched_key_not is not None
                and str(item_pk) in other_dataset.pivot_keys_in
            )
            or (
                self.filter_matched_key
                and str(item_pk) not in other_dataset.pivot_keys_in
            )
        ):
            return False

        return True


class ConflationPrefilters:
    def __init__(
        self, prefilter_a_contain: dict = None, prefilter_b_contain: dict = None
    ) -> None:
        self.prefilter_a_contain = prefilter_a_contain
        self.prefilter_b_contain = prefilter_b_contain

    def _allow_x_contain(self, item: dict, rule_contain: dict = None):
        # if not self.prefilter_a_contain:
        if not rule_contain:
            return True
        elif not item:
            # We assume that if any condition exist to mach, missing
            # tags would fail. However this may not be always the case
            return False

        # count_tries = len(self.prefilter_a_contain.keys)
        okay = None
        # for key, val in self.prefilter_a_contain.items():
        for key, val in rule_contain.items():
            if key not in item:
                continue
            if key in item and val == True:
                okay = True
                continue
            if not isinstance(val, list):
                val = [val]

            _item_term = str(item[key]).lower()

            for alt in val:
                if _item_term.find(alt) > -1:
                    okay = True

        # TODO do some extra attempt here
        if not okay:
            return False
        else:
            return True

        # return True

    def allow_a(self, item: dict):
        return self._allow_x_contain(item, self.prefilter_a_contain)

    def allow_b(self, item: dict):
        return self._allow_x_contain(item, self.prefilter_b_contain)


class ConflationRules:
    def __init__(
        self,
        distance_okay: int = None,
        pivot_key_main: list = None,
        pivot_attr_2: list = None,
        pivot_alias: list = None,
    ) -> None:
        self.distance_okay = distance_okay
        self.pivot_key_main = pivot_key_main
        self.pivot_attr_2 = pivot_attr_2
        self.pivot_alias = pivot_alias


class DatasetInMemory:
    def __init__(
        self,
        alias: str,
        group: str,
        cprefilters: Type["ConflationPrefilters"],
        pivot_key: str = None,
    ) -> None:
        self.alias = alias
        self.group = group
        self.index = -1
        self.cprefilters = cprefilters
        self.pivot_key = str(pivot_key) if pivot_key is not None else None
        # self.is_a = is_a

        # Tuple
        # (coords, props, geometry?)
        # geometry? = only if not already point
        self.items = []
        # self.pk = None
        self.pivot_keys_in = []
        self.pivot_keys_in_duplicate = []

    def add_item(self, item: dict):
        self.index += 1
        # self.items.append(None)

        if (
            not item
            or not isinstance(item, dict)
            or "geometry" not in item
            or not item["geometry"]
            or "coordinates" not in item["geometry"]
        ):
            # Really bad input item (we still count on index)
            self.items.append(False)
            return False

        _properties = item["properties"] if "properties" in item else None

        if (self.group == "A" and not self.cprefilters.allow_a(_properties)) or (
            self.group == "B" and not self.cprefilters.allow_b(_properties)
        ):
            # On this case, we ignore the item
            return False

        if item["geometry"]["type"] != "Point":
            if item["geometry"]["type"] == "Polygon":
                poly = Polygon(item["geometry"]["coordinates"][0])

                coords = (poly.centroid.y, poly.centroid.x)
                props = None
                geometry_original = item["geometry"]
                if (
                    "properties" in item
                    and item["properties"]
                    and len(item["properties"].keys())
                ):
                    props = item["properties"]
                # self.items.append(None)
                # self.items.append((coords, props))

                if (
                    self.pivot_key is not None
                    and props is not None
                    and self.pivot_key in props
                    and props[self.pivot_key]
                ):
                    keyval = str(props[self.pivot_key]).strip()
                    # self.pk = keyval
                    if keyval in self.pivot_keys_in:
                        self.pivot_keys_in_duplicate.append(keyval)
                    else:
                        self.pivot_keys_in.append(keyval)

                self.items.append((coords, props, geometry_original))
            else:
                # For now ignoring non Point features
                self.items.append(None)
        else:
            # Exact point
            coords = (
                item["geometry"]["coordinates"][1],
                item["geometry"]["coordinates"][0],
            )
            props = None

            # Overpass geojson store in "tags" instead of "properties"
            _properties = "tags" if "tags" in item else "properties"

            # if (
            #     "properties" in item
            #     and item["properties"]
            #     and len(item["properties"].keys())
            # ):
            if (
                _properties in item
                and item[_properties]
                and len(item[_properties].keys())
            ):
                props = item[_properties]
            # self.items.append((coords, props))

            if (
                self.pivot_key is not None
                and props is not None
                and self.pivot_key in props
                and props[self.pivot_key]
            ):
                keyval = str(props[self.pivot_key]).strip()
                # self.pk = keyval
                if keyval in self.pivot_keys_in:
                    self.pivot_keys_in_duplicate.append(keyval)
                else:
                    self.pivot_keys_in.append(keyval)

            self.items.append((coords, props, None))

    def item_pk(self, item_index: int) -> str:
        if not self.pivot_key:
            return None

        item_props = self.items[item_index][1]
        if not item_props or self.pivot_key not in item_props:
            return False

        return item_props[self.pivot_key]


class GeojsonCompare:
    """GeojsonCompare

    @TODO optimize for very large files
    """

    def __init__(
        self,
        geodataset_a: str,
        geodataset_b: str,
        strategy: str,
        crules: Type["ConflationRules"],
        cprefilters: Type["ConflationPrefilters"],
        cfilters: Type["ConflationFilters"],
        logger=None,
    ) -> None:
        self.strategy = strategy
        self.distance_okay = crules.distance_okay
        self.crules = crules
        self.cfilters = cfilters
        self.cprefilters = cprefilters
        self.a_is_osm = None
        self.b_is_osm = None

        self.a = self._load_geojson(geodataset_a, "A", "A")
        self.b = self._load_geojson(geodataset_b, "B", "B")
        self.matrix = []
        self.matrix_v2 = []

        self.compute()

        # logger.info(self.summary())
        # pass

    def _load_geojson(self, path: str, alias: str, group: str) -> DatasetInMemory:
        """Load optimized version of GeoJSON++ into memory

        Args:
            path (str): _description_
            alias (str): _description_

        Returns:
            DatasetInMemory
        """
        # print(self.__dict__)
        # print(self.crules.pivot_key_main)
        data = DatasetInMemory(
            alias, group, self.cprefilters, pivot_key=self.crules.pivot_key_main
        )

        with open(path, "r") as file:
            # TODO optimize geojsonl
            jdict = json.load(file)

            # Overpass geojson store in "elements" instead of "features"
            container = "elements" if "elements" in jdict else "features"

            # for feat in jdict["features"]:
            for feat in jdict[container]:
                # print(feat)
                data.add_item(feat)

        return data

    def _short_title(self, properties: dict) -> str:
        if not properties or len(properties.keys()) == 0:
            return ""

        result = ""
        if "nome" in properties:
            result = properties["nome"]
        if "name" in properties:
            result = properties["name"]

        if "ref" in properties:
            result = result + f" ({properties['ref']})"

        return result.strip()

    def compute(self):
        """compute difference of B against A"""
        # for item in self.a.items:

        levenshtein = LevenshteinHelper()

        if len(self.a.items) > 0:
            for index in range(0, len(self.a.items)):
                if not self.a.items[index] or not self.a.items[index][1]:
                    continue

                if (
                    "id" in self.a.items[index][1]
                    and self.a.items[index][1]["id"].startswith(
                        ("node/", "way/", "relation/")
                    )
                ) or (
                    "@id" in self.a.items[index][1]
                    and self.a.items[index][1]["@id"].startswith(
                        ("node/", "way/", "relation/")
                    )
                ):
                    self.a_is_osm = True
                    # print('a_is_osm')
                    break

        if len(self.b.items) > 0:
            for index in range(0, len(self.b.items)):
                # print(self.b.items[index])
                if not self.b.items[index] or not self.b.items[index][1]:
                    continue

                if (
                    "id" in self.b.items[index][1]
                    and self.b.items[index][1]["id"].startswith(
                        ("node/", "way/", "relation/")
                    )
                ) or (
                    "@id" in self.b.items[index][1]
                    and self.b.items[index][1]["@id"].startswith(
                        ("node/", "way/", "relation/")
                    )
                ):
                    self.b_is_osm = True
                    # print('b_is_osm')
                    break

        # if len(self.b.items) > 0:
        #     if "id" in self.b.items[0][1] and self.b.items[0][1]["id"].startswith(
        #         ("node/", "way/", "relation/")
        #     ):
        #         self.b_is_osm = True

        # print(self.a_is_osm, self.b_is_osm)

        for index_a in range(0, len(self.a.items)):
            # print(f"    > teste A i{index_a}", self.a.items[index_a])
            found = False
            # if not self.a.items[index_a]:
            #     self.matrix.append(None)
            # else:
            if self.a.items[index_a]:
                candidates = []
                candidates_aliases = []
                for index_b in range(0, len(self.b.items)):
                    # print("oibb", len(self.b.items))
                    # print('a', self.a.items[index_a])
                    # print('b', self.b.items[index_b][0])

                    # Try perfect match (including tags)
                    if (
                        self.b.items[index_b]
                        and self.a.items[index_a] == self.b.items[index_b]
                    ):
                        self.matrix.append((index_b, MATCH_EXACT, 0, None))
                        self.matrix_v2.append(
                            MatchResult(
                                match_index=index_b,
                                match_dist=0,
                            )
                        )
                        found = True
                        # print("  <<<<< dist zero a")

                    elif (
                        self.b.items[index_b]
                        and self.a.items[index_a][0] == self.b.items[index_b][0]
                    ):
                        # perfect match, except tags (TODO improve this check)
                        self.matrix.append((index_b, MATCH_EXACT, 0, None))
                        self.matrix_v2.append(
                            MatchResult(
                                match_index=index_b,
                                match_dist=0,
                            )
                        )
                        found = True
                        # print("  <<<< dist zero b")

                    # else:
                    elif self.b.items[index_b]:
                        dist = haversine(
                            self.a.items[index_a][0],
                            self.b.items[index_b][0],
                            unit=Unit.METERS,
                        )
                        # print(f"        >> teste A i{index_a} vs B i{index_b}", dist)
                        if dist <= self.distance_okay:
                            # TODO sort by near
                            candidates.append((dist, index_b))
                            # self.matrix.append((index_b, MATCH_NEAR, round(dist, 2)))

                            _aliasgroup = []
                            if self.crules.pivot_alias:
                                for refkey, altkey in self.crules.pivot_alias.items():
                                    if altkey is True:
                                        altkey = refkey
                                if altkey in self.b.items[index_b][1]:
                                    _aliasgroup.append(self.b.items[index_b][1][altkey])
                                else:
                                    # _aliasgroup.append(None)
                                    # _aliasgroup.append('')
                                    pass
                            # print(_aliasgroup)
                            candidates_aliases.append(_aliasgroup)
                            found = True
                        # break

            if found == True and len(candidates) > 0:
                # candidates_aliases = []
                item_aliases = []
                if self.crules.pivot_alias:
                    for refkey, _altkey in self.crules.pivot_alias.items():
                        if refkey in self.a.items[index_a][1]:
                            item_aliases.append(self.a.items[index_a][1][refkey])
                        else:
                            # item_aliases.append(None)
                            item_aliases.append("")
                # print(item_aliases)

                it = ItemMatcher(
                    self.a.items[index_a],
                    item_aliases,
                    candidates,
                    candidates_aliases,
                    self.crules,
                    levenshtein,
                )
                self.matrix.append(it.result())

                _temp = it.result()
                if _temp is None:
                    self.matrix_v2.append(None)
                else:
                    self.matrix_v2.append(
                        MatchResult(
                            match_index=_temp[0],
                            match_dist=_temp[1],
                        )
                    )

                # # pass
                # candidates_sorted = sorted(candidates, key=lambda tup: tup[0])

                # dist = candidates_sorted[0][0]
                # index_b = candidates_sorted[0][1]
                # skiped = None
                # if len(candidates) > 1:
                #     skiped = []
                #     # for i in range(1, len(candidates)):
                #     for i in range(0, len(candidates)):
                #         skiped.append(f"B{candidates[i][1]}")

                # # self.matrix.append((index_b, MATCH_NEAR, round(dist, 2)))
                # self.matrix.append((index_b, MATCH_NEAR, dist, skiped))

            if not found:
                self.matrix.append(None)

    def debug(self):
        print(self.a)
        # print("dataset a", self.a.items)
        print(self.b)
        # print("dataset b", self.b.items)
        print("matrix", self.matrix)

    def diff_geojson_full(self):
        dataobj = {"type": "FeatureCollection", "features": []}

        _repeatcheck = {}

        for index_a in range(0, len(self.a.items)):
            _item_a = self.a.items[index_a]

            if not _item_a:
                # Skip very bad input (invalid geometry)
                continue

            _matrix = self.matrix[index_a]

            final_properties = {}
            pointer = None

            # print("_item_a", _item_a)
            if _item_a[2] is not None:
                final_geometry = _item_a[2]
            else:
                # We assume will be a point
                final_geometry = {
                    "type": "Point",
                    "coordinates": [_item_a[0][1], _item_a[0][0]],
                }

            if _item_a[1] is not None:
                for key, value in _item_a[1].items():
                    # final_properties[f"a.{key}"] = value
                    final_properties[f"{key}"] = value
            # else:
            #     pass

            # if _matrix and self.b.items[_matrix[0]][1]:
            #     # print('_matrix', _matrix)
            #     _item_b = self.b.items[_matrix[0]]

            #     pointer = {
            #         "geometry": {
            #             "type": "LineString",
            #             "coordinates": [
            #                 [_item_a[0][1], _item_a[0][0]],
            #                 [_item_b[0][1], _item_b[0][0]],
            #             ],
            #         },
            #         # "properties": final_properties,
            #         "properties": _item_b[1],
            #         "type": "Feature",
            #         # "_debug": f"_item_a {_item_a}",
            #         # "_original": _item_a,
            #     }

            #     # @TODO fix this
            #     pointer = None

            #     # for key, value in self.b.items[_matrix[0]][1].items():
            #     #     final_properties[f"b.{key}"] = value
            # # else:
            # #     pass

            if _matrix:
                # print("_matrix", _matrix)
                final_properties[f"a->b.distance"] = round(_matrix[2], 2)
                if _matrix[3]:
                    final_properties[f"a->b.near"] = " ".join(_matrix[3])
            else:
                final_properties[f"a->b.distance"] = -1

            if not self.cfilters.dist_ab(final_properties[f"a->b.distance"]):
                continue

            if not self.cfilters.pk_allow(index_a, self.a, self.b):
                continue

            # # Colors inspired by
            # # https://wiki.openstreetmap.org/wiki/OSM_Conflator

            # if final_properties[f"a->b.distance"] == -1:
            #     # 'create': '#11dd11',  # creating a new node
            #     final_properties["action"] = "create"
            #     final_properties["marker-color"] = "#11dd11"
            # elif final_properties[f"a->b.distance"] == 0:
            #     # 'retag':  '#660000',  # cannot delete unmatched feature, changing tags
            #     final_properties["action"] = "retag"
            #     final_properties["marker-color"] = "#660000"
            # else:
            #     # 'move':   '#110055',  # moving an existing node
            #     final_properties["action"] = "move"
            #     final_properties["marker-color"] = "#110055"

            _hash = str(hash(str(final_geometry)))
            if _hash not in _repeatcheck:
                _repeatcheck[_hash] = 0
            else:
                _repeatcheck[_hash] += 1
                if final_geometry["type"] == "Point":
                    final_properties["__coordinates_original"] = final_geometry[
                        "coordinates"
                    ]
                    final_geometry["coordinates"][0] = final_geometry["coordinates"][
                        0
                    ] + (_repeatcheck[_hash] * 0.0001)
                    final_geometry["coordinates"][1] = final_geometry["coordinates"][
                        1
                    ] - (_repeatcheck[_hash] * 0.0001)

            res = {
                "geometry": final_geometry,
                "properties": final_properties,
                "type": "Feature",
                # "_debug": f"_item_a {_item_a}",
                # "_original": _item_a,
            }

            dataobj["features"].append(res)
            if pointer:
                dataobj["features"].append(pointer)
            # dataobj["features"].append(
            #     f"_item_a {_item_a}",
            # )

        return dataobj

    def osmchange(self):
        # @see https://wiki.openstreetmap.org/wiki/OsmChange
        # @see https://wiki.openstreetmap.org/wiki/JOSM_file_format
        if not self.b_is_osm:
            raise SyntaxError(
                "--output-josm-file=file.osm requires Dataset B be an "
                "OpenStreetMap-like geojson. "
                "(id starting with node/N, way/N or relation/N)"
            )

        lines = [
            '<?xml version="1.0" encoding="utf-8"?>',
            "<!-- TEST ONLY, DO NOT UPLOAD! -->",
            f'<osm version="0.6" generator="{PROGRAM} {__VERSION__}" upload="never">',
            f"  <changeset>",
            f'    <tag k="created_by" v="{PROGRAM} {__VERSION__}"/>',
            f"  </changeset>",
        ]

        count = 0

        for index_a in range(0, len(self.a.items)):
            count -= 1
            _item_a = self.a.items[index_a]

            _matrix = self.matrix[index_a]

            n_lat = None
            n_lon = None

            if _item_a[2] is None:
                n_lat = round(_item_a[0][0], 7)
                n_lon = round(_item_a[0][1], 7)
            else:
                lines.append(
                    f"  <!-- {count} ignoring non Point feature suggestion -->"
                )
                continue

            final_properties = {}
            if _item_a[1] is not None:
                for key, value in _item_a[1].items():
                    if value is not False and value is not None and value:
                        final_properties[f"_{key}"] = escape(str(value))

            # @see https://github.com/openstreetmap/osmdbt/issues/29
            #      (topic about order of tags)
            if len(final_properties.keys()) > 0:
                # _kv = sorted(final_properties, key=lambda key: final_properties[key])

                lines.append(
                    f'  <node id="{count}" version="1" lat="{n_lat}" lon="{n_lon}">'
                )
                for key, value in sorted(final_properties.items()):
                    lines.append(f'    <tag k="{key}" v="{value}"/>')
                lines.append(f"  </node>")
            else:
                raise SyntaxError(f"No tags for item {count}. Aborting.")

        # <changeset>
        #     <tag k="source" v="velobike.ru"/>
        #     <tag k="created_by" v="OSM Conflator 1.4.1"/>
        #     <tag k="type" v="import"/>
        # </changeset>

        lines.append("</osm>")
        return "\n".join(lines)

    def summary(self):
        lines = []
        # lines.append("@TODO summary")

        found = 0
        for item in self.matrix:
            if item:
                found += 1

        lines.append(f"A {len(self.a.items)} | {found}")
        lines.append(f"B {len(self.b.items)} | _")

        tabular_out = self.summary_tabular()
        spamwriter = csv.writer(sys.stdout, delimiter="\t")
        for line in tabular_out:
            spamwriter.writerow(line)
        return "\n".join(lines)

    def summary_tabular(self) -> List[list]:
        header = [
            "uid_a",
            "uid_b",
            "id_a",
            "id_b",
            "match_stage",
            "distance_ab",
            "latitude_a",
            "longitude_a",
            "latitude_b",
            "longitude_b",
            "desc_a",
            "desc_b",
            "near_a",
        ]
        data = []

        for index_a in range(0, len(self.a.items)):
            _item_a = self.a.items[index_a]
            if not _item_a:
                # Skip very bad input (invalid geometry)
                continue

            _matrix = self.matrix[index_a]

            # print(_item_a, _matrix)
            # print(_item_a[1], _matrix)
            # # print(_item_a[1])
            # print('aa', _matrix)

            uid_a = f"A{index_a}"
            uid_b = "" if not _matrix else f"B{_matrix[0]}"
            # id_a = "" if not "id" in _item_a else _item_a["id"]
            id_a = (
                "" if not _item_a[1] or not "@id" in _item_a[1] else _item_a[1]["@id"]
            )
            if id_a == "" and _item_a[1] and "id" in _item_a[1]:
                id_a = _item_a[1]["id"]

            id_b = ""
            # if _matrix and self.b.items[_matrix[0]]:
            if (
                _matrix
                and self.b.items[_matrix[0]][1]
                # and "id" in self.b.items[_matrix[0]][1]
            ):
                # print(self.b.items[_matrix[0]][1])
                if "@id" in self.b.items[_matrix[0]][1]:
                    id_b = self.b.items[_matrix[0]][1]["@id"]
                if "id" in self.b.items[_matrix[0]][1]:
                    id_b = self.b.items[_matrix[0]][1]["id"]
                # pass

            match_stage = ""

            # if not _matrix or not "id" in _matrix[1] else _item_a[1]["id"]
            distance_ab = -1 if not _matrix else round(_matrix[2], 2)
            latitude_a = "" if not _item_a else _item_a[0][1]
            longitude_a = "" if not _item_a else _item_a[0][0]
            latitude_b = "" if not _matrix else self.b.items[_matrix[0]][0][1]
            longitude_b = "" if not _matrix else self.b.items[_matrix[0]][0][0]

            # print(self.a.items[index_a])
            desc_a = "" if not _item_a else self._short_title(self.a.items[index_a][1])
            desc_b = (
                "" if not _matrix else self._short_title(self.b.items[_matrix[0]][1])
            )
            # if _matrix:
            #     print(_matrix[3])
            near_a = "" if not _matrix or not _matrix[3] else " ".join(_matrix[3])

            # @TODO implement stages
            if distance_ab > -1:
                match_stage = 5

            # print("index_a", index_a)
            # pass
            data.append(
                [
                    uid_a,
                    uid_b,
                    id_a,
                    id_b,
                    match_stage,
                    distance_ab,
                    latitude_a,
                    longitude_a,
                    latitude_b,
                    longitude_b,
                    desc_a,
                    desc_b,
                    near_a,
                ]
            )

        data.insert(0, header)

        # return data.insert(0, header)
        return data


class ItemMatcher:
    def __init__(
        self,
        item,
        item_aliases: list,
        candidates: list,
        candidates_aliases: list,
        crules: Type["ConflationRules"],
        levenshtein: Type["LevenshteinHelper"],
    ) -> None:
        self.item = item
        # self.candidates = candidates
        ## Improve this part
        self.candidates_valid = None
        self.automacher = None
        self.crules = crules
        self.levenshtein = levenshtein

        self.compute(item_aliases, candidates_aliases, candidates)

    def compute(self, item_aliases, candidates_aliases, candidates):
        # candidates = self.candidates

        # print(self.item, candidates)
        _ext = []
        for idx, item in enumerate(candidates):
            _ext.append((item[0], item[1], item_aliases, candidates_aliases[idx]))
        # candidates_sorted = sorted(candidates, key=lambda tup: tup[0])
        candidates_sorted = sorted(
            _ext, key=lambda tup: Levenshtein.setratio(tup[2], tup[3]), reverse=True
        )

        # Improve this part
        self.candidates_valid = candidates_sorted

        dist = candidates_sorted[0][0]
        index_b = candidates_sorted[0][1]
        skiped = None
        if len(candidates) > 1:
            skiped = []
            # for i in range(1, len(candidates)):
            for i in range(0, len(candidates)):
                skiped.append(f"B{candidates[i][1]}")

        # self.matrix.append((index_b, MATCH_NEAR, round(dist, 2)))
        # self.matrix.append((index_b, MATCH_NEAR, dist, skiped))
        # return (index_b, MATCH_NEAR, dist, skiped)

        self.automacher = (index_b, MATCH_NEAR, dist, skiped)

    def result(self):
        if not self.candidates_valid or len(self.candidates_valid) == 0:
            return None
        if len(self.candidates_valid) == 0:
            return None

        # @TODO order this part better
        # return self.candidates_valid[0]
        if self.automacher:
            return self.automacher
        return None


@dataclasses.dataclass
class MatchResult:
    match_index: int
    match_dist: float


def parse_argument_values(arguments: list, delimiter: str = "||") -> dict:
    if not arguments or len(arguments) == 0 or not arguments[0]:
        return None

    result = {}
    for item in arguments:
        # print('__', item, item.find(delimiter))
        if item.find(delimiter) > -1:
            _key, _val = item.split(delimiter)
            result[_key] = _val
        else:
            result[item] = True

    # print('__f', result)
    return result


def tabular_writer(file_or_stdout: str, data: List[list], delimiter: str = ",") -> None:
    """Write a tabular file

    Args:
        file_or_stdout (str): file or stdout
        data (List[list]): List of lists with data to be outputed
        delimiter (str, optional): Delimiter. Defaults to ",".
    """
    cwriter = csv.writer(file_or_stdout, delimiter=delimiter)
    for line in data:
        cwriter.writerow(line)


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)



import re
import string

# https://pypi.org/project/Levenshtein/
# pip install levenshtein
import Levenshtein

# Record separator
DICTIONARY_SEPARATOR = "␞"


class AttributesEditor:
    """Edit attributes of a dictionary"""

    def __init__(
        self,
        rename_attr: dict = None,
        value_fixed: dict = None,
        value_name_place: list = None,
        normalize_prop: bool = True,
    ) -> None:
        self.rename_attr = rename_attr
        self.value_fixed = value_fixed
        self.value_name_place = value_name_place
        self.normalize_prop = normalize_prop

    def edit(self, item: dict) -> dict:
        result = item

        # print('teste')
        # raise NotImplementedError
        # if self.rename_attr is not None and len(self.rename_attr.keys()) > 0:
        if self.rename_attr is not None:
            for key, val in self.rename_attr.items():
                if key in result:
                    result[val] = result.pop(key)

        # if isinstance(self.value_fixed, list) and len(self.value_fixed) > 0:
        if self.value_fixed:
            for key, val in self.value_fixed.items():
                result[key] = val

        # if isinstance(self.value_name_place, list) and len(self.value_name_place) > 0:
        if self.value_name_place:
            for key in self.value_name_place:
                if key not in result:
                    continue
                _value = _zzz_format_name_place_br(result[key])
                if _value:
                    result[key] = _value
                else:
                    result[key] = ""

        if self.normalize_prop:
            prop_new = {}
            # print(result["properties"])
            for key, val in sorted(result.items()):
                if isinstance(val, str):
                    val = val.strip()
                if not val:
                    continue
                prop_new[key] = val
            result = prop_new

        return result


class DictionaryHelper:
    # @TODO impelement
    def __init__(self) -> None:
        pass


class LevenshteinHelper:
    """LevenshteinHelper is ...
    @see https://maxbachmann.github.io/Levenshtein/levenshtein.html
    pip install levenshtein
    """

    def __init__(
        self, synonymous: dict = None, tokens: dict = None, normalize: bool = True
    ) -> None:
        self.synonymous = synonymous
        self.tokens = tokens
        self.normalize = normalize

    def _apply_synonymous(self, term: str) -> str:
        if not self.synonymous or len(self.synonymous.keys()) == 0:
            return term

        parts = term.split(" ")
        result = []
        for item in parts:
            if item in self.synonymous:
                result.append(self.synonymous[item])
            else:
                result.append(item)

        return " ".join(result)

    def _apply_internal_tokens(self, term: str) -> str:
        if not self.tokens or len(self.tokens.keys()) == 0:
            return term

        parts = term.split(" ")
        result = []
        for item in parts:
            if item in self.tokens:
                result.append(self.tokens[item])
            else:
                result.append(item)

        return " ".join(result)

    def _normalize_term(self, term: str) -> str:
        if not isinstance(term, str):
            term = str(term)
        # if term is not:
        #     return None
        term2 = term.lower()
        term3 = re.sub("\\s\\s+", " ", term2)
        term4 = self._normalize_term_acents(term3)
        term5 = self._apply_synonymous(term4)
        return term5

    def _normalize_term_acents(self, term: str) -> str:
        # Ideally, this should be better than simple replacement
        # @see https://stackoverflow.com/questions/517923
        # @see https://pypi.org/project/Unidecode/
        # pip install Unidecode
        new = term.lower()

        # Obviously incomplete
        new = re.sub(r"[àáâãäå]", "a", new)
        new = re.sub(r"[èéêë]", "e", new)
        new = re.sub(r"[ìíîï]", "i", new)
        new = re.sub(r"[òóôõö]", "o", new)
        return new

    def get_term_tokenized(self, term: str) -> str:
        """get_term_tokenized Return a term expanded and normalized
        Args:
            term (str): The input term

        Returns:
            str: the result

        >>> synonymous = {'7': "sete"}
        >>> tokens = {'avenida': "<PPREFIX>", 'rua': "<PPREFIX>"}
        >>> lh = LevenshteinHelper(synonymous=synonymous)
        >>> lh.get_term_tokenized('rua sete de setembro')
        'rua sete de setembro'
        """
        # return Levenshtein.distance("teste", "testeee")

        # if self.normalize:
        #     term = self._normalize_term(term)
        term = self._normalize_term(term)
        term2 = self._apply_synonymous(term)

        return term2

    def get_term_tokenized_internal(self, term: str) -> str:
        """get_term_tokenized Return a term expanded and normalized
        Args:
            term (str): The input term

        Returns:
            str: the result

        >>> synonymous = {'7': "sete"}
        >>> tokens = {'avenida': "<PPREFIX>", 'rua': "<PPREFIX>"}
        >>> lh = LevenshteinHelper(synonymous=synonymous)
        >>> lh.get_term_tokenized('TESTÉ')
        'teste'

        >>> lh.get_term_tokenized_internal('rua sete de setembro')
        'rua sete de setembro'
        """
        term = self._normalize_term(term)
        term2 = self._apply_synonymous(term)
        term3 = self._apply_internal_tokens(term2)

        return term3

    def distance(self, ref_term: str, alt_term: str) -> int:
        """distance

        Args:
            ref_term (str): reference term
            alt_term (str): alternative term

        Returns:
            int: distance value

        >>> synonymous = {'7': "sete"}
        >>> tokens = {'avenida': "<PPREFIX>", 'rua': "<PPREFIX>"}
        >>> lh = LevenshteinHelper(synonymous=synonymous, tokens=tokens)

        #>>> lh.distance('Rua 7 de Setembro', 'rua sete de setembro')
        #0

        >>> lh.distance('Avenida 7 de Setembro', 'rua sete de setembro')
        0
        """
        if self.normalize:
            ref_term = self.get_term_tokenized_internal(ref_term)
            alt_term = self.get_term_tokenized_internal(alt_term)

        # print(self.normalize)
        # print(ref_term)
        # print(alt_term)

        return Levenshtein.distance(ref_term, alt_term)

    def sorted(self, ref_terms: list, alt_terms: list, alt_original: list) -> list:
        # @TODO sort the alt_original by ref_terms VS alt_terms
        return alt_original

    def get_sorted_alternatives(self, ref_term: str, alt_terms: list) -> list:
        alt_terms_sorted_index = range(0, len(alt_terms))

        return alt_terms_sorted_index


# def parse_argument_values(arguments: list, delimiter: str = "||") -> dict:
def parse_argument_values(
    arguments: list, delimiter: str = "|||", delimiter2: str = "||"
) -> dict:
    if not arguments or len(arguments) == 0 or not arguments[0]:
        return None

    result = {}
    for item in arguments:
        # print('__', item, item.find(delimiter))
        if item.find(delimiter) > -1:
            _key, _val = item.split(delimiter)
            if _val.find(delimiter2) > -1:
                _val = item.split(_val)
            result[_key] = _val
        else:
            result[item] = True

    # print('__f', result)
    return result


# @TODO refactor this code and allow localization of values
def _zzz_format_name_place_br(value: str):
    if not value or not isinstance(value, str):
        return value

    term = string.capwords(value.strip())
    term2 = re.sub("\\s\\s+", " ", term)

    # @TODO deal with Do Da De

    term2 = term2.replace(" Da ", " da ")
    term2 = term2.replace(" De ", " de ")
    term2 = term2.replace(" Do ", " do ")

    return term2


def _zzz_format_phone_br(value: str):
    if not value:
        return False

    if value.startswith("+"):
        return value

    # @TODO deal with more than one number

    if value.startswith("(") and value.find(")") > -1:
        _num_com_ddd = "".join(filter(str.isdigit, value))

        _regiao = _num_com_ddd[0:2]
        _num_loc = _num_com_ddd[2:]
        _num_loc_p2 = _num_loc[-4:]
        _num_loc_p1 = _num_loc.replace(_num_loc_p2, "")
        # return "+55 " + _regiao + ' ' + _num_com_ddd[2:]
        return "+55 " + _regiao + " " + _num_loc_p1 + " " + _num_loc_p2

    # if value.isnumeric():
    #     if len(value) == 8:
    #         return re.sub(r"(\d{5})(\d{3})", r"\1-\2", value)
    return False



# pip install pytest-testinfra
# pytest


# def test_wikitext_001_jsonld():
#     assert True


# def test_wikitext_003_jsonld():
#     assert True


def test_csv2excel_cmd(host):
    cmd = host.run("csv2excel --help")
    assert cmd.rc == 0


def test_csv2geojson_cmd(host):
    cmd = host.run("csv2geojson --help")
    assert cmd.rc == 0


def test_geojsondiff_cmd(host):
    cmd = host.run("geojsondiff --help")
    assert cmd.rc == 0


# def test_nginx_running_and_enabled(host):
#     nginx = host.service("nginx")
#     assert nginx.is_running
#     assert nginx.is_enabled



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  geojsonedit.py
#
#         USAGE:  geojsonedit --help
#                 ./src/gisconflation/geojsonedit.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v0.5.0
#       CREATED:  2023-04-28 14:17 BRT
# ==============================================================================

import argparse

# import csv
# import dataclasses
import json
import sys

# import logging
from typing import List, Type

from .util import AttributesEditor, parse_argument_values

# from haversine import haversine, Unit

# # from shapely.geometry import Polygon, Point
# from shapely.geometry import Polygon
# from xml.sax.saxutils import escape


__VERSION__ = "0.5.0"

PROGRAM = "geojsonedit"
DESCRIPTION = """
------------------------------------------------------------------------------
GeoJSON / GeoJSONSeq command line editor

------------------------------------------------------------------------------
""".format(
    __file__
)

# https://www.rfc-editor.org/rfc/rfc7946
# The GeoJSON Format
# https://www.rfc-editor.org/rfc/rfc8142
# GeoJSON Text Sequences

# __EPILOGUM__ = ""
__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
    {0} --help

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer

# MATCH_EXACT = 1
# MATCH_NEAR = 3


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args

        Args:
            hxl_output (bool, optional): _description_. Defaults to True.
        """
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        # @TODO maybe implement GeoJSONSeq
        parser.add_argument("input", help="Input GeoJSON / GeoJSONSeq. Use - for stdin")
        # parser.add_argument("geodataset_b", help="GeoJSON dataset 'B'")

        parser.add_argument(
            "--format-output",
            help="Path to output file",
            dest="format_output",
            default="auto",
            required=False,
            choices=["auto", "geojson", "geojsonseq"],
            nargs="?",
        )

        edit = parser.add_argument_group("Change properties from each item")

        edit.add_argument(
            "--rename-attribute",
            help="Rename attribute (if exist). "
            "Use '||' to divide the source key and target key. "
            "Accept multiple values. "
            "Example: "
            "--rename-attribute='NAME|||name' --rename-attribute='CITY|||addr:city'",
            dest="rename_attr",
            nargs="?",
            action="append",
        )

        # Old versions of csv2geojson uses only one |
        edit.add_argument(
            "--value-fixed",
            help="Define a fixed string for every value of a column, "
            "For multiple, use multiple times this parameter. "
            "Source vs destiny column must be divided by |||. "
            "Example: <[ --value-fixed='source|||BR:DATASUS' ]>",
            dest="value_fixed",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        edit.add_argument(
            "--value-norm-name-place",
            help="Column to normalize value (strategy: name of place). "
            "Accept multiple options. Example: "
            "--value-norm-name-place='name' --value-norm-name-place='alt_name'",
            dest="value_name_place",
            nargs="?",
            # type=lambda x: x.split("||"),
            action="append",
            default=None,
        )

        filter = parser.add_argument_group("Options for filter input items completely")

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # logger = logging.getLogger()
        # logger.setLevel(logging.INFO)
        # if pyargs.outlog:
        #     fh = logging.FileHandler(pyargs.outlog)
        #     logger.addHandler(fh)
        # else:
        #     ch = logging.StreamHandler()
        #     logger.addHandler(ch)

        normalize_prop = True
        skip_invalid_geometry = True

        # raise ValueError(parse_argument_values(pyargs.value_fixed))
        # print(parse_argument_values(pyargs.value_fixed))

        gitem = GeoJSONItemEditor(
            rename_attr=parse_argument_values(pyargs.rename_attr),
            value_fixed=parse_argument_values(pyargs.value_fixed),
            value_name_place=pyargs.value_name_place,
            normalize_prop=normalize_prop,
            skip_invalid_geometry=skip_invalid_geometry,
        )
        gedit = GeoJSONFileEditor(pyargs.input, gitem)
        gedit.output()

        # geodiff.debug()
        return self.EXIT_OK


# ./src/gisconflation/geojsonedit.py tests/data/geojson-simple.geojson
# cat tests/data/geojson-simple.geojson | ./src/gisconflation/geojsonedit.py -
# ./src/gisconflation/geojsonedit.py tests/data/geojson-seq.geojsonl
class GeoJSONFileEditor:
    """GeoJSONEditor

    @TODO implement read line-by-line large files (in special GeoJSONSeq)
    """

    def __init__(self, input_ptr: str, gitem: Type["GeoJSONItemEditor"]) -> None:
        self.gitem = gitem
        self.inputdata = self._loader_temp(input_ptr)
        # pass

    def _loader_temp(self, input_ptr: str):
        # @TODO make this efficient for very large files
        data = None
        if input_ptr == "-":
            temp = []
            for line in sys.stdin:
                # temp.append(line.strip())
                temp.append(line.rstrip())
                # temp.append(line)

            data = "\n".join(temp)
        else:
            with open(input_ptr, "r") as _file:
                data = _file.read()
            # pass

        return data

    def output(self):
        json_data = json.loads(self.inputdata)
        if not json_data or "features" not in json_data:
            raise SyntaxError("bad input data")

        result = {"features": []}
        for item in json_data["features"]:
            edited_item = self.gitem.edit(item)
            if edited_item is not False:
                result["features"].append(edited_item)

        # print(self.inputdata)
        # print(json_data)

        # @TODO keep other metadata at top level, if exist
        print('{"type": "FeatureCollection", "features": [')
        count = len(json_data["features"])
        for item in json_data["features"]:
            count -= 1
            line_str = json.dumps(item, ensure_ascii=False)
            if count > 0:
                line_str += ","
            print(line_str)
        print("]}")

        # print(json.dumps(result, ensure_ascii=False))


class GeoJSONItemEditor:
    def __init__(
        self,
        rename_attr: dict = None,
        value_fixed: dict = None,
        value_name_place: list = None,
        normalize_prop: bool = True,
        skip_invalid_geometry: bool = True,
    ) -> None:
        self.rename_attr = rename_attr
        self.normalize_prop = normalize_prop
        self.value_fixed = value_fixed
        self.value_name_place = value_name_place
        self.skip_invalid_geometry = skip_invalid_geometry

        self._attr_editor = AttributesEditor(
            rename_attr=rename_attr,
            value_fixed=value_fixed,
            value_name_place=value_name_place,
            normalize_prop=normalize_prop,
        )

        # print(self.rename_attr)
        # pass

    def edit(self, item: dict):
        # return item
        # print(item)
        result = item

        if self.skip_invalid_geometry:
            if (
                not item
                or not isinstance(item, dict)
                or "geometry" not in item
                or not item["geometry"]
                or "coordinates" not in item["geometry"]
                # or not item["geometry"]["coordinates"]
            ):
                # @TODO make better checks, like if is Type=Point, etc
                return False

        if "properties" in result:
            result["properties"] = self._attr_editor.edit(result["properties"])

            # if self.rename_attr is not None and len(self.rename_attr.keys()) > 0:
            #     for key, val in self.rename_attr.items():
            #         if key in result["properties"]:
            #             result["properties"][val] = result["properties"].pop(key)

            # if self.normalize_prop:
            #     prop_new = {}
            #     # print(result["properties"])
            #     for key, val in sorted(result["properties"].items()):
            #         if isinstance(val, str):
            #             val = val.strip()
            #         if not val:
            #             continue
            #         prop_new[key] = val
            #     result["properties"] = prop_new
        return result


# def parse_argument_values(arguments: list, delimiter: str = "||") -> dict:
#     if not arguments or len(arguments) == 0 or not arguments[0]:
#         return None

#     result = {}
#     for item in arguments:
#         # print('__', item, item.find(delimiter))
#         if item.find(delimiter) > -1:
#             _key, _val = item.split(delimiter)
#             result[_key] = _val
#         else:
#             result[item] = True

#     # print('__f', result)
#     return result


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  dictionarymerger.py
#
#         USAGE:  ./scripts/dictionarymerger.py
#                 ./scripts/dictionarymerger.py --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v1.0.0
#       CREATED:  2023-05-06 22:24 BRT started, based on dictionarybuilder.py
#      REVISION:  --
# ==============================================================================

import argparse
import csv
# import re
import sys


__VERSION__ = "1.0.0"
PROGRAM = "dictionarymerger"
DESCRIPTION = """
------------------------------------------------------------------------------
For 2 or more output dictionaries from dictionarybuilder, merge them.

Simple strategy is *append* (e.g. similar to unix cat command, but sort again
the values and check inconsistencies).

[Next part NOT implemented yet]

The additional strategy is *transpose* value from dictionary A directly on
result of dictoryary B. Example:

    Input 1: X -> Y
    Input 2: Y -> Z
    Output:  X -> Z

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------

    {0} --help

------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

# Record separator
DICTIONARY_SEPARATOR = "␞"

STDIN = sys.stdin.buffer


class Cli:
    """Main CLI parser"""

    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        # @see https://stackoverflow.com/questions/5373474/multiple-positional-arguments-with-python-and-argparse
        # @TODO deal with both simple contatenation and transposition
        #       do not need to implement on this tool advanced input normalization
        parser.add_argument("input", help="path to CSV file on disk. Use - for stdin")

        parser.add_argument(
            "--input-append-file",
            help="The input delimiter",
            dest="in_append_files",
            default=None,
            action="append",
            required=False,
            nargs="?",
        )

        parser.add_argument(
            "--ignore-warnings",
            help="Ignore some errors (duplicated key / ambiguous results)",
            dest="ignore_warnings",
            action="store_true",
        )

        parser.add_argument(
            "--logfile",
            help="Path to a file to log warnings and other information",
            dest="logfile",
            nargs="?",
        )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        # with open(pyargs.input, "r", encoding=pyargs.in_encoding) if len(

        outdict = {}

        f_warnings = sys.stderr
        if pyargs.logfile:
            f_logfile = open(pyargs.logfile, "w", encoding="utf-8")
            f_warnings = f_logfile

        with open(pyargs.input, "r") if len(pyargs.input) > 1 else sys.stdin as csvfile:
            reader = csv.reader(csvfile, delimiter=DICTIONARY_SEPARATOR)

            for row in reader:
                # print(row)
                outdict[row[0]] = row[1]

            outdict_sorted = dict(sorted(outdict.items()))

            writer = csv.writer(
                sys.stdout, delimiter=DICTIONARY_SEPARATOR, quoting=csv.QUOTE_MINIMAL
            )

        if pyargs.in_append_files:
            for dfile in pyargs.in_append_files:
                # @TODO deal with duplicates
                with open(dfile, "r") as csvfile:
                    reader = csv.reader(csvfile, delimiter=DICTIONARY_SEPARATOR)
                    for row in reader:
                        # print(dfile)
                        # print(row)
                        _k = row[0]
                        _v = row[1]
                        if (
                            _k in outdict
                            and outdict[_k] != _v
                            and not pyargs.ignore_warnings
                        ):
                            print(f"{_k} repeated", file=f_warnings)
                        outdict[_k] = _v

        for key, value in outdict_sorted.items():
            writer.writerow([key, value])

        if pyargs.logfile:
            f_logfile.close()

        return self.EXIT_OK


def _file_to_dict(file: str) -> dict:
    pass


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_2600 = Cli()
    args = cli_2600.make_args()
    # pyargs.print_help()

    # args.execute_cli(args)
    cli_2600.execute_cli(args)



#!/usr/bin/env python3
# ==============================================================================
#
#          FILE:  overpassql2osmf.py
#
#         USAGE:  overpassql2osmf --help
#
#   DESCRIPTION:  ---
#
#       OPTIONS:  ---
#
#  REQUIREMENTS:  - python3
#                   - pip install pandas (https://pandas.pydata.org/)
#          BUGS:  ---
#         NOTES:  ---
#       AUTHORS:  Emerson Rocha <rocha[at]ieee.org>
# COLLABORATORS:  ---
#
#       COMPANY:  EticaAI
#       LICENSE:  GNU Affero General Public License v3.0 or later
#                 SPDX-License-Identifier: AGPL-3.0-or-later
#       VERSION:  v0.2.0
#       CREATED:  2023-04-26 01:04 BRT
#      REVISION:  ---
# ==============================================================================

# ./src/gisconflation/overpassql2osmf.py --help
# ./src/gisconflation/overpassql2osmf.py tests/data/cnes.overpassql > tests/temp/cnes.osm.json
# overpassql2osmf tests/data/cnes.overpassql > tests/temp/cnes.osm.json
# overpassql2osmf tests/data/cnes.overpassql | osmf2geojson > tests/temp/cnes.osm.geojson
# ./src/gisconflation/overpassql2osmf.py '[out:csv(::id,::type,"name")]; area[name="Bonn"]; nwr(area)[railway=station]; out;' > tests/temp/bonn.osm.csv


# overpassql2osmf tests/data/cnes.overpassql > tests/temp/cnes.osm
# osmium sort tests/temp/cnes.osm --output-format osm > tests/temp/cnes-sorted.osm
# osmium export tests/temp/cnes-sorted.osm --output-format geojson > tests/temp/cnes.osm.geojson

# osmf2geojson tests/temp/cnes-sorted.osm > tests/temp/cnes.osm.geojson

import os
import sys
import argparse

import requests

# import pandas as pd

__VERSION__ = "0.2.0"
PROGRAM = "overpassql2osmf"
DESCRIPTION = """
------------------------------------------------------------------------------
{0} v{1} convert Overpass Query Language to am OSM File (XML)

------------------------------------------------------------------------------
""".format(
    PROGRAM, __VERSION__
)

__EPILOGUM__ = """
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
Input file contains Overpass query . . . . . . . . . . . . . . . . . . . . . .
    {0} tests/data/cnes.overpassql > tests/temp/cnes.osm

CSV Output (requires special Overpass query) . . . . . . . . . . . . . . . . .

    {0} '[out:csv(::id,::type,"name")]; area[name="Bonn"]; nwr(area)[railway=station]; out;' \
> tests/temp/bonn.osm.csv
------------------------------------------------------------------------------
                            EXEMPLŌRUM GRATIĀ
------------------------------------------------------------------------------
""".format(
    PROGRAM
)

STDIN = sys.stdin.buffer
OVERPASS_INTERPRETER = os.getenv(
    "OVERPASS_INTERPRETER", "https://overpass-api.de/api/interpreter"
)


class Cli:
    def __init__(self):
        """
        Constructs all the necessary attributes for the Cli object.
        """
        self.pyargs = None
        self.EXIT_OK = 0
        self.EXIT_ERROR = 1
        self.EXIT_SYNTAX = 2

    def make_args(self):
        """make_args"""
        parser = argparse.ArgumentParser(
            prog=PROGRAM,
            description=DESCRIPTION,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog=__EPILOGUM__,
        )

        parser.add_argument(
            "input_query", help="Input OverpassQL query or path to .overpassql file"
        )
        # parser.add_argument("output_file", help="Output XLSX")

        # # @see https://stackoverflow.com/questions/41669690/how-to-overcome-the-limit-of-hyperlinks-in-excel
        # parser.add_argument(
        #     "--hiperlink",
        #     help="Add hiperlink to column names if they have non-empty value. "
        #     "alredy are mached with each other. "
        #     "Use '||' to deparate colum name from the URL. "
        #     "Add {value} as part of the URL for placeholder of external link. "
        #     "Accept multiple values. "
        #     "Example: "
        #     "--pivot-key-main='colum_name||http://example.com/page/{value}'",
        #     dest="hiperlink",
        #     nargs="?",
        #     action="append",
        # )

        return parser.parse_args()

    def execute_cli(self, pyargs, stdin=STDIN, stdout=sys.stdout, stderr=sys.stderr):
        query = pyargs.input_query
        if pyargs.input_query.endswith(".overpassql"):
            with open(pyargs.input_query, "r") as file:
                query = file.read()

        # print(query)
        # print('')
        # print('')

        result_str = overpassql2osmf(
            query,
        )
        print(result_str)
        # with open(pyargs.output_file, "w") as file:
        #     file.write(result_str)

        return self.EXIT_OK


def overpassql2osmf(input_query: str):
    """overpassql2osmf
    @see https://xlsxwriter.readthedocs.io/example_pandas_autofilter.html

    Args:
        input_query (str): Input CSV file
        output_file (str): Output XLSX
    """
    payload = {"data": input_query}
    r = requests.post(OVERPASS_INTERPRETER, data=payload)

    return r.text


def parse_argument_values(arguments: list, delimiter: str = "||") -> dict:
    if not arguments or len(arguments) == 0 or not arguments[0]:
        return None

    result = {}
    for item in arguments:
        if item.find(delimiter) > -1:
            _key, _val = item.split(delimiter)
            result[_key] = _val
        else:
            result[_key] = True
    return result


def exec_from_console_scripts():
    main = Cli()
    args = main.make_args()
    main.execute_cli(args)


if __name__ == "__main__":
    cli_main = Cli()
    args = cli_main.make_args()
    cli_main.execute_cli(args)
